import pubchempy as pcp
import json
# import time
from typing import Union
import pandas as pd
from ..utils.helper import Singleton

class PubChemServer(metaclass=Singleton):

    def __init__(self) -> None:
        self.properties = {value: key for key, value in pcp.PROPERTY_MAP.items()}

    def get_columns(self, selected_columns: list[str]):
        return [pcp.PROPERTY_MAP[col] for col in selected_columns]

    def get_compounds(self, comp: Union[list[str], str], selected_columns: list[str], domain: str='compound', namespace: str='cid'):
        
        operation = 'property/' + ','.join(selected_columns)

        try:
            result = pcp.request(
                identifier=comp,
                domain=domain,
                namespace=namespace,
                operation=operation,
                output='JSON',
                # searchtype=None
            ).read()
        except pcp.PubChemHTTPError as e:
            raise e
        
        result = json.loads(result.decode())

        result = result['PropertyTable']['Properties']

        compounds = pd.DataFrame.from_dict(result)

        return compounds
    
    def get_synonyms(self, comp: Union[list[str], str], domain: str='compound', namespace: str='cid'):

        try:
            result = pcp.request(
                identifier=comp,
                domain=domain,
                namespace=namespace,
                operation='synonyms',
                output='JSON',
                # searchtype=None
            ).read()
        except pcp.PubChemHTTPError as e:
            raise e
        
        result = json.loads(result.decode())

        result = result['InformationList']['Information']

        compounds = pd.DataFrame.from_dict(result)
        compounds['Synonym'] = compounds['Synonym'].apply(lambda x: ','.join(x) if isinstance(x, list) else x)

        return compounds



import pandas as pd
import numpy as np
import time
import io
import requests
from ..servers.PubchemServer import PubChemServer
from ..servers.BiomartServer import BiomartServer
from ..data_manager import *
from ..utils.helper import generate_subsets
from .Database import Database

class Stitch(Database):
    
    def __init__(self, connection=None, database=None):
        # if not connection and not database:
        #     raise ValueError('Either SQL connection or database should be not None')
        if database is not None:
            self.data_manager = LocalManager(database)
        else:
            self.data_manager = SQLManager(connection, 'Stitch')

    def interactions(self, input_comp: pd.DataFrame, set_stereo: bool=True):
        """
        Retrieves proteins from Stitch database interacting with compound passed as input.

        Steps
        -----
        - Creates specific and unspecific identifiers from input CID.
        - Finds matches on the database, based on the set_stereo value \\
        Constraint:
            - Only experimental evidence
        - Links stitch to STRING to obtain updated protein ids and symbols.
        - Uses Biomart to obtain proteins info (modified with data from original DTC database) to return,
        searching with ensembl peptide id.
        
        Parameters
        ----------
        DTC_data : DataFrame
            Dataframe containing all DTC database info
        input_comp : DataFrame
            Dataframe of input compounds from which interacting proteins are found
        set_stereo : bool
            value to determine which kind of stereochemistry to consider for the input compound:
            - True (default) - specific stereochemistry
            - False - non-specific stereochemistry \\ 
            True ensures exact match to input_id

        Returns
        -------
        DataFrame
            Dataframe of interacting proteins, containing the following values:
            - entrez 
            - gene_type 
            - hgnc_symbol 
            - description 
            - datasource (Stitch) 
            - pchembl_value (NaN)
        String
            A statement string describing the outcome of the database search
        DataFrame
            Raw Dataframe containing all Stitch info about the input compound
        """

        columns = ['entrez','gene_type','hgnc_symbol','description','pchembl_value','datasource']
        # Create an empty DataFrame with the specified columns
        sttch_act = pd.DataFrame(columns=columns)
        # Create an empty Pubchem activity DataFrame
        sttch_raw = pd.DataFrame(columns=columns)
        # Drop null values of CID, to make sure the first line of the dataframe has the CID
        input_comp = input_comp.dropna(subset=['cid']).reset_index(drop=True)
        # Check if there are any input compounds remaining 
        if len(input_comp) > 0:
            # Find STITCH interactions for chemical by PubChem CID
            # Need to convert CID into stitch id, need a set number of leading zeroes (len = 8)
            CID = str(input_comp['cid'][0])
            # STITCH adds prefix CIDs for specific stereochemistry
            sttch1 = 'CIDs' + CID.zfill(8)
            # STITCH adds prefix CIDm for unspecific stereochemistry
            sttch2 = 'CIDm' + CID.zfill(8)
            
            # Merges all stereochemistries together
            if not set_stereo: 
                stitch_stereo = self.data_manager.retrieve_raw_data('chemical', sttch1)
                sttch_flat = self.data_manager.retrieve_raw_data('chemical', sttch2)

                sttch_raw = pd.concat([stitch_stereo,sttch_flat]).reset_index(drop=True)
            # Only uses the select input stereochemistry
            else: 
                sttch_raw = self.data_manager.retrieve_raw_data('chemical', sttch1)
        
            # Check if at least one match has been found
            if len(sttch_raw) > 0:
                # Filter STITCH to experimental evidence only
                sttch_act = sttch_raw.loc[sttch_raw['experimental'] > 0].reset_index(drop=True)
                # STITCH has no activity values so cannot calculate pChEMBL
                # The input STITCH data should already be filtered to human only interactions
                
                # Check if there is at least one interaction left
                if len(sttch_act) > 0:
                    # Linking Stitch to STRING to obtain updated protein IDs 
                    
                    # sttch_act['String_symbol'] = 'String Conversion Failure'
                    sttch_act['String_id'] = sttch_act['protein']

                    ids = list(sttch_act['protein'].unique())
                    
                    # Extract string ids using Stitch API
                    genes = self._extract_ids(ids)

                    if len(genes) > 0:
                        genes = genes[['output', 'symbol']].rename(columns={'output':'String_id',
                                                                            'symbol': 'String_symbol'})
                        sttch_act = pd.merge(sttch_act, genes, on='String_id', how='left')
                        
                    # Unify gene identifiers by Converting STITCH with biomart
                    ensembl = BiomartServer()
                    # Stitch uses ensembl peptide id
                    input_type = 'ensembl_peptide_id'
                    attributes = ['ensembl_peptide_id', 'entrezgene_id', 'gene_biotype', 'hgnc_symbol', 'description']
                    names = ['ensembl_peptide_id','entrez','gene_type','hgnc_symbol','description']

                    stitch_targets = pd.DataFrame(columns=names)    

                    # Removes the first 5 characters from each id
                    sttch_act.loc[:, ['String_id']] = sttch_act['String_id'].str[5:]
            
                    input_genes = list(sttch_act['String_id'])
                    
                    stitch_targets=ensembl.subset_search(input_type, input_genes, attributes, names)

                    # For each compound, assign specific biomart column values to the ones from the original Stitch database
                    for index, row in sttch_act.iterrows():
                        S1 = stitch_targets.loc[stitch_targets['ensembl_peptide_id']==row['String_id']]
                        if len(S1) > 0:
                            sttch_act.loc[index,'entrez'] = S1['entrez'].iloc[0]
                            sttch_act.loc[index,'gene_type'] = S1['gene_type'].iloc[0]
                            sttch_act.loc[index,'hgnc_symbol'] = S1['hgnc_symbol'].iloc[0]
                            sttch_act.loc[index,'description'] = S1['description'].iloc[0]
                        else:
                            sttch_act.loc[index,'entrez'] = None
                            sttch_act.loc[index,'gene_type'] = None
                            sttch_act.loc[index,'hgnc_symbol'] = None
                            sttch_act.loc[index,'description'] = None
                            Des = 'Failed to convert the gene ID'                   
                    sttch_act['datasource'] = 'Stitch'
                    sttch_act['pchembl_value'] = np.nan
                    statement = 'completed'
                else:
                    statement = 'Filter reduced interactions to 0'
            else:
                statement = 'No interaction data'
        else:
            statement = 'Input compound doesn\'t contain CID'
        return sttch_act, statement, sttch_raw


    def compounds(self, input_protein: pd.DataFrame, set_stereo: bool=True):
        """
        Retrieves compounds from Stitch database interacting with proteins passed as input.

        Steps
        -----
        - Filters stitch database to find compounds interacting with input proteins: \\
        Constraints:
            - stereochemistry = specific
        - Uses Pubchempy to obtain compounds info to return, searching with CID.
        
        Parameters
        ----------
        input_protein : DataFrame
            Dataframe of input proteins from which interacting compound are found
        stitch_data : DataFrame
            Dataframe containing all stitch database info        
        set_stereo : bool
            value to determine which kind of stereochemistry to consider for the input compound:
            - True (default) - specific stereochemistry
            - False - non-specific stereochemistry \\ 
            True ensures exact match to input_id

        Returns
        -------
        DataFrame
            Dataframe of interacting compounds, containing the following values:
            - inchi
            - inchikey
            - isomeric smiles
            - iupac name
            - datasource (Stitch)
            - pchembl value (NaN)
            - notes (NaN)
        """

        columns = ['inchi','inchikey','isomeric_smiles','iupac_name','datasource','pchembl_value']
        # Create an empty DataFrame with the specified columns
        stitch_c1 = pd.DataFrame(columns=columns)
        stitch_raw = pd.DataFrame()
        # Drop duplicated and null values of uniprot and ensembl peptide id
        input_protein = input_protein.dropna(subset=['uniprot', 'ensembl_peptide_id'])
        input_protein = input_protein.drop_duplicates(subset='uniprot')
        input_protein = input_protein.drop_duplicates(subset='ensembl_peptide_id').reset_index(drop=True)
        # Check if there are any input proteins remaining   
        if len(input_protein) > 0:
            input_protein_id = input_protein['ensembl_peptide_id'][0]
            input_protein_id = '9606.' + str(input_protein_id)
            # Remove homo sapiens identifier from the protein id column
            stitch_raw = self.data_manager.retrieve_raw_data('protein', input_protein_id)
            stitch_raw['protein'] = stitch_raw['protein'].str.replace('9606.', '', regex=True)
            # stitch_raw = stitch_data.loc[stitch_data['protein']==input_protein_id]

            if len(stitch_raw) > 0:
                if set_stereo:
                    # Select only specific (stereochemistry) compounds, not non specific
                    stitch_act = stitch_raw.loc[stitch_raw['chemical'].str.contains("s", case=False, na=False)].copy()
                # Remove CIDs and CIDm strings from chemical column
                stitch_act['CID'] = stitch_act['chemical'].str.replace(r'^(CIDs|CIDm)', '', regex=True).str.lstrip('0')
                stitch_act = stitch_act.dropna(subset=['CID'])
                # Select only non empty cids and experimental values > 0
                stitch_act = stitch_act.loc[(~stitch_act['CID'].eq('')) & 
                                        (stitch_act['experimental'] > 0)].reset_index(drop=True)
                # Check if there are any compounds remaining
                if len(stitch_act) > 0:

                    pc = PubChemServer()
                    # Retrieve compounds using cids
                    stitch_c1 = self._pubchem_search_cid(stitch_act, columns, pc)

                    if len(stitch_c1) > 0:
                        stitch_c1.loc[:, 'pchembl_value'] = np.nan        
                        stitch_c1.loc[:, 'datasource'] = 'Stitch'
                        stitch_c1.loc[:, 'notes'] = np.nan
                        statement = 'completed'
                    else:
                        statement = 'Compounds not found using PubChem'
                else:
                    statement = 'Filter reduced interactions to 0'
            else:
                statement = 'No interaction data'
        else:
            statement = 'Input protein doesn\'t contain ensembl id'  

        return stitch_c1, statement, stitch_raw

    
    def _extract_ids(self, ids):

        names = ['input','value','output','taxid','species','symbol','description']
        params={
                # Protein list
                "identifiers" : None, 
                # species NCBI identifier (Human is being used)
                "species" : 9606, 
                # Only one (best) identifier per input protein
                "limit" : 1, 
                # See input identifiers in the output
                "echo_query" : 1, 
                # App name
                "caller_identity" : "NEU" 
            }
        # Construct URL and Call STRING
        string_api_url="https://version-11-5.string-db.org/api"
        output_format="tsv-no-header"
        method="get_string_ids"
        request_url = '/'.join([string_api_url,output_format,method])

        if len(ids) > 250:
            genes = pd.DataFrame(columns=names)
            for start, end in generate_subsets(len(ids), 250):
                subset = ids[start:end]
                params['identifiers'] = "\r".join([str(gene) for gene in subset])
                try:
                    response = requests.post(request_url, data=params)
                    # Read and parse the results
                    st = pd.read_csv(io.StringIO(response.text), sep='\t', header=None, 
                                        names=['input','value','output','taxid','species','symbol','description'])
                    genes = pd.concat([genes, st])
                except:
                    continue
                time.sleep(0.5)
        else:
            params['identifiers'] = "\r".join([str(gene) for gene in ids])
            try:
                response = requests.post(request_url, data=params)
                genes = pd.read_csv(io.StringIO(response.text), sep='\t', header=None, 
                            names=['input','value','output','taxid','species','symbol','description'])
            except:
                genes = pd.DataFrame(columns=names)
        genes['output'] = genes['output'].astype(str)
        return genes



from .DataManager import DataManager
import pandas as pd

class LocalManager(DataManager):

    def retrieve_raw_data(self, filter_column, filter_value, **kwargs):

        if self.data is None:
            raise ValueError('Database not available')
        
        if type(filter_value) == list:
            # Filter data, create copy and reset index
            raw_data = self.data[self.data[filter_column].isin(filter_value)].copy().reset_index(drop=True)
        else:
            # Filter data, create copy and reset index
            raw_data = self.data[self.data[filter_column] == filter_value].copy().reset_index(drop=True)
        
        return raw_data


from .identifiers import compound_identifiers, protein_identifiers
from .helper import generate_subsets, Singleton

__all__ = ['generate_subsets', 'Singleton', 'compound_identifiers', 'protein_identifiers']


import numpy as np
import pandas as pd
from ..utils.identifiers import compound_identifiers
from ..databases import *
from .Pipeline import Pipeline

class Comp2Prot(Pipeline):

    def _update_args(self, pChEMBL_thres, stitch_stereo, otp_biblio, dtc_mutated, dc_extra, chembl_ids):
        self.database_args = {
            'pc': (pChEMBL_thres,),
            'chembl': (chembl_ids, pChEMBL_thres,),
            'bdb': (pChEMBL_thres,),
            'stitch': (stitch_stereo,),
            'ctd': (),
            'dtc': (chembl_ids, dtc_mutated, pChEMBL_thres,),
            'otp': (chembl_ids, otp_biblio,),
            'dc': (dc_extra, pChEMBL_thres,),
            'db': (),
        }       


    # Calls functions to collect data and merges all the data from the various sources together  
    # The parameters are:
    #    - input_id - the compound id
    #    - pChEMBL_thresh - the minimum interaction pChEMBL value required to be added to the output file
    #    - stitch_stereo - to select whether to consider the specific compound stereochemistry or group all stereoisomers interactions from STITCH
    #    - otp_biblio - to select whether to include the *bibliography* data from OTP. This parameter is only available for Comp2Prot as OTP provides only known drug interactions for proteins
    #    - dtc_mutated - to select whether also to consider interactions with mutated target proteins from DTC
    #    - dc_extra - to select whether to include possibly non-Homo sapiens interactions

    def comp_interactions(self, input_id, pChEMBL_thres=0, stitch_stereo=True, 
                          otp_biblio=False, dtc_mutated=False, dc_extra=False):
        
        # Run interaction select with all databases selected
        comp_tar, state = self.comp_interactions_select(input_id, pChEMBL_thres=pChEMBL_thres, 
                                                    stitch_stereo=stitch_stereo, otp_biblio=otp_biblio, 
                                                    dtc_mutated=dtc_mutated, dc_extra=dc_extra)
        return comp_tar, state

    # Calls functions to collect data and merges all the data from the selected sources together
    # The parameters are:
    #    - input_id - the compound id
    #    - selected_dbs - underscore-separated string containing the names of the databases which are selected
    #    - pChEMBL_thresh - the minimum interaction pChEMBL value required to be added to the output file
    #    - stitch_stereo - to select whether to consider the specific compound stereochemistry or group all stereoisomers interactions from STITCH
    #    - otp_biblio - to select whether to include the *bibliography* data from OTP. This parameter is only available for Comp2Prot as OTP provides only known drug interactions for proteins
    #    - dtc_mutated - to select whether also to consider interactions with mutated target proteins from DTC
    #    - dc_extra - to select whether to include possibly non-Homo sapiens interactions

    def comp_interactions_select(self, input_id, selected_dbs='pc_chembl_bdb_stitch_ctd_dtc_otp_dc_db', 
                                 pChEMBL_thres=0, stitch_stereo=True, otp_biblio=False, dtc_mutated=False, dc_extra=False):
        #pChEMBL_thres default 0: Will only interactions without activity data in sources with activity data present
        #stitch_stereo default True: True=stereo specific False=non-specific stereochemistry (True ensures exact match to input_id)
        #otp_biblio default False: False=mechanism data only True=mechanism and bibliography (False ensures real interactions, avoids keyword search matches within abstracts that can lead to false associations that are not real interactions)
        
        chembl_ids = []
        self._update_args(pChEMBL_thres, stitch_stereo, otp_biblio, dtc_mutated, dc_extra, chembl_ids)

        main_columns = ['entrez','gene_type','hgnc_symbol','description','pchembl_value','datasource']
        comp_tar = pd.DataFrame(columns=main_columns)
        #Collect all the data through the other functions
        comp_ids = compound_identifiers(input_id) #find comp_ids
        #Outputs the brief statement regarding the processing of each source 
        states = pd.DataFrame(columns=['comp'])
        states.loc[0, 'comp'] = input_id

        if len(comp_ids) > 0:
            tar_all = pd.DataFrame(columns=main_columns)
            db_list = selected_dbs.split('_')
            for name, db in self.databases.items():
                if name in db_list:
                    # Perform database search
                    result, state, _ = db.interactions(comp_ids, *self.database_args.get(name, ()))
                    if len(result) > 0:
                        # Keep only wanted columns
                        result = result.loc[:, main_columns]
                        # Concatenate the source
                        tar_all = pd.concat([tar_all, result])
                    print(f'{name} done!')
                else:
                    state = f'Did not run {name}'
                # Add statement for this source
                states.loc[0, name] = state
            
            comp_tar = self._postprocess_databases(tar_all)

            if len(comp_tar) > 0:
                # Add std compound ids to output
                std_ids = comp_ids[['isomeric_smiles', 'inchi', 'inchikey', 'iupac_name']]
                # Select first nonNa value from each column
                std_ids = std_ids.apply(lambda col: std_ids[col.name].dropna().iloc[0] if not std_ids[col.name].dropna().empty else None)
                # Assign to each row the standard ids
                comp_tar = comp_tar.assign(**std_ids)       
                # Add input id to the results
                comp_tar['input_id'] = input_id

                # Reorder columns
                comp_tar = comp_tar[['input_id', 'inchi', 'inchikey', 'isomeric_smiles', 'iupac_name', 
                                    'entrez', 'gene_type', 'hgnc_symbol', 'description',
                                    'pchembl_count', 'ave_pchembl', 'std_pchembl', 'src_count', 'pubchem',
                                    'chembl', 'bindingdb', 'stitch', 'ctd', 'dtc', 'otp', 'drugcentral',
                                    'drugbank']]

        return comp_tar, states
    

    def _postprocess_databases(self, tar_all):

        # Remove non-protein coding interactions
        tar_all = tar_all[tar_all['gene_type']=='protein_coding']
        # Remove proteins with no Symbol or Entrez
        tar_all = tar_all.dropna(subset=['hgnc_symbol', 'entrez']).reset_index(drop=True)  
        # Make a list of the unique targets for the compound
        tar_list = tar_all['hgnc_symbol'].unique() 
        
        #Create an output dataframe 
        comp_tar=pd.DataFrame(columns=['entrez','hgnc_symbol','description','gene_type','pchembl_count',
                                       'ave_pchembl','src_count'] + 
                                       [source.lower() for source in self.sources])
        
        for index, target in enumerate(tar_list):
            tar = tar_all.loc[tar_all['hgnc_symbol'] == target] #Select target
            
            # Copy the biomart data to the output dataframe
            comp_tar.loc[index,'entrez'] = int(tar['entrez'][tar.index[0]])
            comp_tar.loc[index,'hgnc_symbol'] = tar['hgnc_symbol'][tar.index[0]]
            comp_tar.loc[index,'description'] = tar['description'][tar.index[0]]
            comp_tar.loc[index,'gene_type'] = tar['gene_type'][tar.index[0]]
            
            comp_tar = self._aggregate_pchembl(comp_tar, index, tar)
            
            # Sets a source binary matrix to the dataframe to easily filter output later
            comp_tar.loc[index, 'src_count'] = len(tar['datasource'].unique())
            # Sets a source binary matrix to the dataframe to easily filter output later
            for source in self.sources:
                comp_tar.loc[index, source.lower()] = 1 if source in tar['datasource'].unique() else 0


        return comp_tar



import pandas as pd
import numpy as np
from ..servers.PubchemServer import PubChemServer
from ..servers.BiomartServer import BiomartServer
from .Database import Database
from ..data_manager import *

class CTD(Database):

    def __init__(self, connection=None, database=None):
        # if not connection and not database:
        #     raise ValueError('Either SQL connection or database should be not None')
        if database is not None:
            self.data_manager = LocalManager(database)
        else:
            self.data_manager = SQLManager(connection, 'CTD')

    def _filter_database(self, CTD_raw: pd.DataFrame):
                        # Filter to Homo Sapiens only
        CTD_act = CTD_raw.loc[(CTD_raw['OrganismID'] == 9606) &
                            # Only select ligand-protein binding
                            (CTD_raw['InteractionActions'].str.contains('affects\^binding')) &
                            # Only protein interactions are collected
                            (CTD_raw['GeneForms']==('protein'))].drop_duplicates().reset_index(drop=True)
        return CTD_act

    def interactions(self, input_comp: pd.DataFrame):
        """
        Retrieves proteins from CTD database interacting with compound passed as input.

        Steps
        -----
        - Finds matches for all synonyms of the compound passed as input
        - Finds all proteins interacting with input compound \\
        Constraints:
            - Only Homo Sapiens interactions
            - Only ligand-protein binding
            - Only protein interactions
        - Uses Biomart to obtain proteins info (modified with data from from original CTD database) to return,
        searching with GeneID.
        
        Parameters
        ----------
        CTD_data : DataFrame
            Dataframe containing all CTD database info
        input_comp : DataFrame
            Dataframe of input compounds from which interacting proteins are found

        Returns
        -------
        DataFrame
        DataFrame
            Dataframe of interacting proteins, containing the following values:
            - entrez 
            - gene_type 
            - hgnc_symbol 
            - description 
            - datasource (CTD) 
            - pchembl_value (NaN)
        String
            A statement string describing the outcome of the database search
        DataFrame
            Raw Dataframe containing all CTD info about the input compound
        """

        columns = ['entrez','gene_type','hgnc_symbol','description','pchembl_value','datasource']
        # Create an empty DataFrame with the specified columns
        CTD_act = pd.DataFrame(columns=columns)
        CTD_raw = pd.DataFrame()
        input_comp = input_comp.dropna(subset=['cid'])
        if len(input_comp) > 0:
            input_comp_id = input_comp['cid'][0]

            CTD_raw = self.data_manager.retrieve_raw_data('CID', input_comp_id)
            # CTD_raw = CTD_data[CTD_data['cid'] == input_comp_id].reset_index(drop=True)

            # Check if at least one match has been found
            if len(CTD_raw) > 0:
                # Ensure a binding type interaction occurs
                CTD_act = self._filter_database(CTD_raw)
                
                # Check if at least one interaction has been found
                if len(CTD_act) > 0:
                    # Unify gene identifiers by Converting CTD with Biomart
                    ensembl = BiomartServer()
                    attributes = ['entrezgene_id', 'gene_biotype', 'hgnc_symbol', 'description']
                    names = ['entrez','gene_type','hgnc_symbol','description']
                    # Use entrez for conversion
                    input_type='entrezgene_id' 
                    
                    ctd_targets=pd.DataFrame(columns=names)
                    
                    input_genes=list(CTD_act['GeneID'])
                    
                    ctd_targets=ensembl.subset_search(input_type, input_genes, attributes, names)

                    # For each compound, assign specific biomart column values to the ones from the original CTD database
                    for index, row in CTD_act.iterrows():
                        S1 = ctd_targets.loc[ctd_targets['entrez']==int(row['GeneID'])]
                        if len(S1) > 0:
                            CTD_act.loc[index,'entrez'] = S1['entrez'].iloc[0]
                            CTD_act.loc[index,'gene_type'] = S1['gene_type'].iloc[0]
                            CTD_act.loc[index,'hgnc_symbol'] = S1['hgnc_symbol'].iloc[0]
                            CTD_act.loc[index,'description'] = S1['description'].iloc[0]
                        else:
                            CTD_act.loc[index, 'entrez'] = None
                            CTD_act.loc[index, 'gene_type'] = None
                            CTD_act.loc[index, 'hgnc_symbol'] = None
                            CTD_act.loc[index, 'description'] = None  
                            Des='Failed to convert the gene ID'
                    CTD_act['datasource'] = 'CTD'
                    # No activity values in database so no pChEMBL score calculated
                    CTD_act['pchembl_value'] = np.nan
                    statement = 'completed'
                else:
                    statement = 'Filter reduced interactions to 0'
            else:
                statement = 'No interaction data'
        else:
            statement = 'Compound doesn\'t have any cid'

        return CTD_act, statement, CTD_raw
    
    
    def compounds(self, input_protein: pd.DataFrame):
        """
        Retrieves compounds from CTD database interacting with proteins passed as input.

        Steps
        -----
        - Filters CTD database to obtain compounds interacting with input proteins \\
        Constraints:
            - Organism = homo sapiens
            - InteractionActions contains 'affects\^binding'
            - GeneForms = protein 
        - Uses Pubchempy to obtain compounds info to return, searching with CID.

        Parameters
        ----------
        input_protein : DataFrame
            Dataframe of input proteins from which interacting compound are found
        CTD_data : DataFrame
            Dataframe containing all CTD database info

        Returns
        -------
        DataFrame
            Dataframe of interacting compounds, containing the following values:
            - inchi
            - inchikey
            - isomeric smiles
            - iupac name
            - datasource (CTD)
            - pchembl value (NaN)
            - notes (NaN)
        """

        columns = ['inchi','inchikey','isomeric_smiles','iupac_name','datasource','pchembl_value']
        # Create an empty DataFrame with the specified columns
        ctd_c1 = pd.DataFrame(columns=columns)
        ctd_raw = pd.DataFrame()
        input_protein = input_protein.dropna(subset=['entrez'])
        input_protein = input_protein.drop_duplicates(subset='entrez').reset_index(drop=True) 
        # Check if there are any input proteins remaining
        if len(input_protein) > 0:
            # Filter database
            input_protein_id = int(input_protein['entrez'].iloc[0])
            ctd_raw = self.data_manager.retrieve_raw_data('GeneID', input_protein_id)

            # Check if at least one match has been found
            if len(ctd_raw) > 0:

                ctd_act = self._filter_database(ctd_raw)

                if len(ctd_act) > 0:
                    pc = PubChemServer()
                    # Retrieve compounds using cids
                    ctd_c1 = self._pubchem_search_cid(ctd_raw, columns, pc)
                    
                    if len(ctd_c1) > 0:
                        ctd_c1['datasource'] = 'CTD'
                        ctd_c1['pchembl_value'] = np.nan
                        ctd_c1.loc[:, 'notes'] = np.nan 
                        statement = 'completed'
                    else:
                        statement = 'Compounds not found using PubChem'
                else:
                    statement = 'Filter reduced interactions to 0'
            else:
                statement = 'No interaction data'
        else:
            statement = 'Input protein doesn\'t contain entrez'  

        return ctd_c1, statement, ctd_raw   


import numpy as np
import pandas as pd
from ..servers.BiomartServer import BiomartServer
from ..servers.PubchemServer import PubChemServer
from .Database import Database
from ..data_manager import *
import time


class DrugCentral(Database):

    def __init__(self, connection=None, database=None):
        # if not connection and not database:
        #     raise ValueError('Either SQL connection or database should be not None')
        if database is not None:
            self.data_manager = LocalManager(database)
        else:
            self.data_manager = SQLManager(connection, 'DC')

    def _filter_database(self, dc_raw: pd.DataFrame, dc_extra: bool):

        dc_filt = dc_raw.dropna(subset=['ACTION_TYPE'])
        dc_filt = dc_filt.loc[(dc_filt['ORGANISM']=='Homo sapiens') &
                             (dc_filt['ACT_TYPE'].isin(['IC50', 'Ki', 'EC50', 'Kd', 'AC50'])) &
                             (dc_filt['ACT_SOURCE'] != 'UNKNOWN')]\
                                .drop_duplicates(ignore_index=True).copy()
        if not dc_extra:
            invalid_action_types = ['PHARMACOLOGICAL CHAPERONE', 'RELEASING AGENT']
            invalid_classes = ['CD molecules', 'RNA', 'Unclassified', 'Viral envelope protein', 'Polyprotein']
            dc_filt = dc_filt.loc[(~dc_filt['ACTION_TYPE'].isin(invalid_action_types)) & 
                                  (~dc_filt['TARGET_CLASS'].isin(invalid_classes))].copy()
        return dc_filt
    

    def _compute_pchembl(self, dc_dat: pd.DataFrame, pChEMBL_thres: float):

        dc_dat['pchembl_value'] = dc_dat['ACT_VALUE'].apply(lambda x: -np.log10(x / 1e9))
        dc_act = dc_dat.rename(columns={'ACT_TYPE': 'notes'})

        # Filter interactions by pChEMBL value
        dc_act = dc_act.loc[dc_act['pchembl_value'] > pChEMBL_thres].reset_index(drop=True) 

        return dc_act

    def interactions(self, input_comp: pd.DataFrame, dc_extra: bool=False, pChEMBL_thres: float=0):
        """
        Retrieves proteins from DrugCentral database interacting with compound passed as input.

        Steps
        -----
        - Filters dc database to obtain proteins interacting with input compound \\
        Constraints:
            - Only Homo sapiens interactions
            - Activity Type must be one of the following: IC50, Ki, EC50, Kd, AC50
            - Action Type must not be: PHARMACOLOGICAL CHAPERONE, RELEASING AGENT
            - Target Class must not be: CD molecules, RNA, Unclassified, Viral envelope protein, Polyprotein
        - Uses activity column to compute pchembl values.
        - Uses Biomart to obtain proteins info (modified with data from original dc database) to return,
        searching with UniProt ID.
        
        Parameters
        ----------
        input_comp : DataFrame
            Dataframe of input compound data from which interacting proteins are found
        dc_extra: bool
            bool to select whether to include possibly non-Homo sapiens interactions
        pChEMBL_thres : float
            minimum pChEMBL value necessary for interaction to be considered valid
            
        Returns
        -------
        DataFrame
            Dataframe of interacting proteins, containing the following values: \\
            entrez, gene_type, hgnc_symbol, description, datasource (DrugCentral), pchembl_value
        String
            A statement string describing the outcome of the database search
        DataFrame
            Raw Dataframe containing all DrugCentral info about the input compound
        """

        columns = ['entrez','gene_type','hgnc_symbol','description','pchembl_value','datasource']
        # Create an empty DataFrame with the specified columns
        dc_act = pd.DataFrame(columns=columns)
        dc_raw = pd.DataFrame()
        # Drop null values of inchikey, to make sure the first line of the dataframe has the inchi key 
        input_comp = input_comp.dropna(subset=['inchikey']).reset_index(drop=True)
        # Check if there are any input compounds remaining  
        if len(input_comp) > 0:
            # Select compound from BindingDB data based on inchi key
            input_comp_id = input_comp['inchikey'][0]
            dc_raw = self.data_manager.retrieve_raw_data('InChIKey', input_comp_id)
            if len(dc_raw) > 0:
                # Filter database
                dc_filt = self._filter_database(dc_raw, dc_extra)
                dc_filt.drop(['SMILES','InChI'], axis=1, inplace=True)
                # Compute pchembl values
                dc_act = self._compute_pchembl(dc_filt, pChEMBL_thres)
                
                if len(dc_act) > 0:
                    # Unify gene identifiers by Converting BindingDB with biomart
                    ensembl = BiomartServer()
                    # BindingDB uses Uniprot IDs
                    input_type='uniprotswissprot' 
                    attributes = ['uniprotswissprot', 'entrezgene_id', 'gene_biotype', 'hgnc_symbol', 'description']
                    names = ['uniprot','entrez','gene_type','hgnc_symbol','description']
                    dc_targets = pd.DataFrame(columns=names)
                    
                    input_genes = list(dc_act['ACCESSION'])
                    
                    dc_targets = ensembl.subset_search(input_type, input_genes, attributes, names)
    
                    # For each protein, assign specific biomart column values to the ones from the original dc database
                    for index, row in dc_act.iterrows():
                        # Find the compound using uniprot
                        S1 = dc_targets.loc[dc_targets['uniprot'] == row['ACCESSION']]
                        if len(S1) > 0:
                            dc_act.loc[index, 'entrez'] = S1['entrez'].iloc[0]
                            dc_act.loc[index, 'gene_type'] = S1['gene_type'].iloc[0]
                            dc_act.loc[index, 'hgnc_symbol'] = S1['hgnc_symbol'].iloc[0]
                            dc_act.loc[index, 'description'] = S1['description'].iloc[0]
                        else:
                            dc_act.loc[index, 'entrez'] = None
                            dc_act.loc[index, 'gene_type'] = None
                            dc_act.loc[index, 'hgnc_symbol'] = None
                            dc_act.loc[index, 'description'] = None
                            Des='Failed to convert the gene ID'
                    dc_act['datasource'] = 'DrugCentral'
                    statement = 'completed'
                else:
                    statement = 'Filter reduced interactions to 0'
            else:
                statement = 'No interaction data'
        else:
            statement = 'Input compound doesn\'t contain inchi key'

        return dc_act, statement, dc_raw
    

    def compounds(self, input_protein: pd.DataFrame, dc_extra: bool=False, pChEMBL_thres: float=0):
        """
        Retrieves compounds from DrugCentral database interacting with proteins passed as input.

        Steps
        -----
        - Filters dc database to obtain compounds interacting with input proteins:
        Constraints:
            - Only Homo sapiens interactions
            - Activity Type must be one of the following: IC50, Ki, EC50, Kd, AC50
            - Action Type must not be: PHARMACOLOGICAL CHAPERONE, RELEASING AGENT
            - Target Class must not be: CD molecules, RNA, Unclassified, Viral envelope protein, Polyprotein
        - Uses activity column to compute pchembl values. \\
        - Uses Pubchempy to obtain compounds info (modified with data from original dc database) to return,
        searching with InChiKey
        
        Parameters
        ----------
        input_protein : DataFrame
            Dataframe of input proteins from which interacting compound are found
        dc_extra: bool
            bool to select whether to include possibly non-Homo sapiens interactions
        pChEMBL_thres : float
            minimum pChEMBL value necessary for interaction to be considered valid       

        Returns
        -------
        DataFrame
            Dataframe of interacting compounds, containing the following values: \\
            inchi, inchikey, isomeric_smiles, iupac_name, datasource (DrugCentral), pchembl_value, notes (activity type)
        """

        columns = ['inchi','inchikey','isomeric_smiles','iupac_name','datasource','pchembl_value']
        # Create an empty DataFrame with the specified columns
        dc_c1 = pd.DataFrame(columns=columns)
        dc_raw = pd.DataFrame()
        # Drop duplicated and null values of uniprot, to make sure the first line of the dataframe has the uniprot ID
        input_protein= input_protein.dropna(subset=['uniprot'])
        input_protein= input_protein.drop_duplicates(subset='uniprot').reset_index(drop=True) 
        # Check if there are any input proteins remaining       
        if len(input_protein) > 0:
            input_protein_id=input_protein['uniprot'].iloc[0]
            # Select only compounds interacting with the input protein
            dc_raw = self.data_manager.retrieve_raw_data('ACCESSION', input_protein_id)
            if len(dc_raw):
                # Filter database
                dc_c = self._filter_database(dc_raw, dc_extra)
                
                # Compute pchembl values
                dc_act = self._compute_pchembl(dc_c, pChEMBL_thres)
                # Check if at least one interacting compound has been found
                if len(dc_act) > 0:
                    pc = PubChemServer()

                    if 'CID' in dc_act.columns:
                        # Retrieve compounds using cids
                        compounds = self._pubchem_search_cid(dc_act, columns, pc)
                        
                        if len(compounds) > 0:
                            # Add additional values from activity dataframe
                            dc_info = dc_act[['CID', 'ACCESSION', 'TARGET_CLASS', 'ACT_COMMENT',
                                                'pchembl_value', 'notes']].\
                                        rename(columns={'ACCESSION': 'uniprotid'}).\
                                                        drop_duplicates()
                            
                            dc_c1 = pd.merge(compounds, dc_info, on='CID', how='left')
                    else:
                        # Filter only columns from pubchempy to return
                        selected_columns = pc.get_columns(columns[:-2])
                        compounds = pd.DataFrame()
                        ids = list(dc_act['InChIKey'].unique())
                        for id in ids:
                            try:
                                comp = pc.get_compounds(id, selected_columns, namespace='inchikey')
                                compounds = pd.concat([compounds, comp])
                            except:
                                None
                            time.sleep(0.5)
                        # Check if at least a match has been found
                        if len(compounds) > 0:
                            dc_c1 = compounds.rename(columns=pc.properties)

                            # Add additional values from activity dataframe
                            dc_info = dc_act[['InChIKey', 'ACCESSION', 'TARGET_CLASS', 'ACT_COMMENT',
                                                'pchembl_value', 'notes']].\
                                        rename(columns={'InChIKey': 'inchikey', 
                                                        'ACCESSION': 'uniprotid'}).\
                                                        drop_duplicates()
                            
                            dc_c1 = pd.merge(dc_c1, dc_info, on='inchikey', how='left')

                    if len(dc_c1) > 0:
                        dc_c1['datasource'] = 'DrugCentral'
                        statement = 'completed'
                    else:
                        statement = 'Compounds not found using PubChem'
                else:
                    statement = 'Filter reduced interactions to 0'
            else:
                statement = 'No interaction data'
        else:
            statement = 'Input protein doesn\'t contain UniProt ID'
        return dc_c1, statement, dc_raw



import pandas as pd
import numpy as np
from ..servers.BiomartServer import BiomartServer
from ..servers.ChEMBLServer import ChEMBLServer as chembl
from ..servers.PubchemServer import PubChemServer
from .Database import Database
from ..data_manager import *


class DTC(Database):

    def __init__(self, connection=None, database=None):
        # if not connection and not database:
        #     raise ValueError('Either SQL connection or database should be not None')
        if database is not None:
            self.data_manager = LocalManager(database)
        else:
            self.data_manager = SQLManager(connection, 'DTC')


    def _filter_database(self, DTC_raw: pd.DataFrame):
        # Filter DTC_data to measurements and units for calculation of pChEMBL
        valid_standard_types = ['IC50','KI','EC50','KD','Kd','ED50','XC50','AC50','Ki','ED50','CC50','AVERAGEIC50',
                            'ACTIVITYEC50','SC50','AD50','DC50','XI50','GC50','GI50','LOGEC50','-LOGIC50','PIC50',
                            'LOGIC50','LOGKD','LOGKI','LOGKI','logIC50','LOGIC50','-LOGKD','PEC50','LOGEC50',
                            'PIC50(CALC)','LOG(10^6/IC50)','-LOGED50','LOG1/IC50','PED50','LOGKD','LOG1/KD',
                            'LOG(1/IC50)','1/KI','PXC50']
        valid_unit_types = ['NM','M','/NM','NMOL/L','MG.KG-1','UG.ML-1','UMOL.KG-1','P.P.M.','PPM',
                        'UG KG-1','MG KG-1','MG/ML','NG ML-1',"10'-4UMOL/L",'UG ML-1','MM','NMOL/MG',
                        "10'13NM","10'-10M","10'8NM",'10^5/M','MOL/G',"10'-8M",'NM G-1','M-1',"10'-5M",
                        "10'7NM",'/UM',"10'-7M","10'-9M","10'-9L/MOL","10'-10L/MOL",'PM','UM L-1','UG/ML']
        
                            # Keep only valid standard types
        DTC_filt = DTC_raw.loc[DTC_raw['standard_type'].isin(valid_standard_types) & 
                            # Keep only valid standard units
                            DTC_raw['standard_units'].isin(valid_unit_types) &
                            # Remove entries with no measurement
                            DTC_raw['standard_value'].notnull() &
                            # Remove entries without target info
                            DTC_raw['target_id'].notnull() &
                            # Remove entries without compound id
                            DTC_raw['compound_id'].notnull()
                            ].drop_duplicates().copy().reset_index(drop=True)  
        return DTC_filt
        

    def _standardize_database(self, DTC_raw: pd.DataFrame, dtc_mutated: bool, pChEMBL_thres: float):

                    # Only use interactions with reported sources
        DTC_act = DTC_raw[(DTC_raw['doc_type'].notnull()) & 
                        # Remove inconclusive and undetermined activities
                        (~DTC_raw['activity_comment'].isin(['inconclusive', 'Not Active']))].reset_index(drop=True)
        if not dtc_mutated:
            # Remove all mutated data
            DTC_act = DTC_act[(DTC_act['wildtype_or_mutant'] != 'mutated') & 
                        (DTC_act['mutation_info'].isnull())].reset_index(drop=True)
        # Converts all the measured values into nM for the calculation of pChEMBL
        DTC_act = self._standard_converter(DTC_act)
        # Convert infinite values to nan
        DTC_act['standardized_val'].replace([np.inf, -np.inf], np.nan, inplace=True)
        # Remove empty values
        DTC_act = DTC_act[DTC_act['standardized_val'].notnull()]
        # Remove non-numeric characters from the Activity Value (things like > and <)
        #DTC_act['standardized_val']=DTC_act['standardized_val'].str.replace(r'[^0-9\.]','',regex=True)
        # Convert strings into numeric values 
        DTC_act['standardized_val'] = DTC_act['standardized_val'].apply(pd.to_numeric,errors='coerce')
        # Generate pChEMBL values
        DTC_act['pchembl_value'] = -np.log10(DTC_act['standardized_val']/1e9)  
        # Filter interactions by pChEMBL value
        DTC_act=DTC_act.loc[DTC_act['pchembl_value'] > pChEMBL_thres].reset_index(drop=True) 

        return DTC_act
            

    def interactions(self, input_comp: pd.DataFrame, chembl_ids: list, dtc_mutated: bool=False, pChEMBL_thres: float=0):
        """
        Retrieves proteins from DTC database interacting with compound passed as input.

        Steps
        -----
        - Filters input database from unnecessary data \\
        Constraints:
            - Only valid standard types 
            - Only valid standard units
            - Entries with measurement only
            - Entiries with target info only \\
        - Finds matches on the database for all compounds ids passed as input and applies a second filter \\
        Constraints:
            - Only interactions with reported sources
            - No inconclusive and undetermined activities
            - No mutated data \\
        - Applies a standard conversion for all values required to compute pChembl, removing non valid values.
        - Uses Biomart to obtain proteins info (modified with data from original DTC database) to return,
        searching with uniprotswissprot.
        
        Parameters
        ----------
        DTC_data : DataFrame
            Dataframe containing all DTC database info
        input_comp : DataFrame
            Dataframe of input compounds from which interacting proteins are found
        chembl_ids : list
            list of chembl ids for the input compounds
        dtc_mutated: bool
            bool to select whether to included mutated targets interactions
        pChEMBL_thres : float
            minimum pChEMBL value necessary for interaction to be considered valid
        
        Returns
        -------
        DataFrame
            Dataframe of interacting proteins, containing the following values: \\
            entrez, gene_type, hgnc_symbol, description, datasource (DTC), pchembl_value (nan)
        String
            A statement string describing the outcome of the database search
        DataFrame
            Raw Dataframe containing all DTC info about the input compound
        """

        columns = ['entrez','gene_type','hgnc_symbol','description','pchembl_value','datasource']
        # Create an empty DataFrame with the specified columns
        DTC_act = pd.DataFrame(columns=columns)
        DTC_raw = pd.DataFrame(columns=columns)
        
        # Check if chembl_ids has already been computed from the input compound, otherwise do so
        if not chembl_ids:
            chembl_ids.extend(chembl.identify_chembl_ids(input_comp))
        
        if len(chembl_ids) > 0 and None not in chembl_ids:
            # Filter DTC by compound of interest using ChEMBL IDs
            DTC_raw = self.data_manager.retrieve_raw_data('compound_id', chembl_ids)
            # DTC_raw = DTC_data.loc[DTC_data['compound_id'].isin(chembl_ids)].reset_index(drop=True)
            
            # Check if at least one match has been found
            if len(DTC_raw) > 0:

                DTC_filt = self._filter_database(DTC_raw)

                DTC_filt['molecular_weight'] = input_comp['molecular_weight'].iloc[0]

                # Filter database
                DTC_act = self._standardize_database(DTC_filt, dtc_mutated, pChEMBL_thres)
                
                # Note: DTC has no tax_id or species column, can only filter for human via biomart conversion
                if len(DTC_act) > 0:
                    # Unify gene identifiers by Converting DTC with biomart
                    ensembl = BiomartServer()
                    attributes = ['uniprotswissprot', 'entrezgene_id', 'gene_biotype', 'hgnc_symbol', 'description']
                    names = ['uniprot','entrez','gene_type','hgnc_symbol','description']
                    # Search by uniprot id match to biomart
                    input_type = 'uniprotswissprot'
                    
                    dtc_targets = pd.DataFrame(columns=names)
                    
                    input_genes = list(DTC_act['target_id'])
                    
                    dtc_targets = ensembl.subset_search(input_type, input_genes, attributes, names)

                    # For each compound, assign specific biomart column values to the ones from the original DTC database
                    for index, row in DTC_act.iterrows():
                        S1 = dtc_targets.loc[dtc_targets['uniprot']==row['target_id']]
                        if len(S1) > 0:
                            DTC_act.loc[index,'entrez'] = S1['entrez'].iloc[0]
                            DTC_act.loc[index,'gene_type'] = S1['gene_type'].iloc[0]
                            DTC_act.loc[index,'hgnc_symbol'] = S1['hgnc_symbol'].iloc[0]
                            DTC_act.loc[index,'description'] = S1['description'].iloc[0]
                        else:
                            DTC_act.loc[index, 'entrez'] = None
                            DTC_act.loc[index, 'gene_type'] = None
                            DTC_act.loc[index, 'hgnc_symbol'] = None
                            DTC_act.loc[index, 'description'] = None  
                            Des='Failed to convert the gene ID'
                    DTC_act['datasource']='DTC'
                    statement='completed'
                else:
                    statement='Filter reduced interactions to 0'
            else:
                statement='No interaction data'
        else:
            statement = 'No ChEMBL ids found'

        return DTC_act, statement, DTC_raw
    

    def compounds(self, input_protein: pd.DataFrame, dtc_mutated: bool=False, pChEMBL_thres: float=0):
        """
        Retrieves compounds from DTC database interacting with proteins passed as input.

        Steps
        -----
        - Retrieves gene data from DTC database to find compounds interacting with gene
        Constraints:
            - Standard type and units not null
            - pChEMBL value > threshold
        - Uses Pubchempy to retrieve the compound info (with additional data from chembl) to return, 
        searching with compound ID or with inchi (obtained from chembl API) if the first search is unsuccessful.
        - Applies a standard conversion for all values required to compute pChembl.
        
        Parameters
        ----------
        input_protein : DataFrame
            Dataframe of input proteins from which interacting compound are found
        DTC_data : DataFrame
            Dataframe containing all DTC database info
        dtc_mutated: bool
            bool to select whether to included mutated targets interactions
        pChEMBL_thres : integer
            Threshold for pChEMBL value to be valid

        Returns
        -------
        DataFrame
            Dataframe of interacting compounds, containing the following values: \\
            inchi, inchikey, isomeric_smiles, iupac_name, datasource (DTC), pchembl_value, notes (nan)
        """

        columns = ['inchi','inchikey','isomeric_smiles','iupac_name','molecular_weight','datasource','pchembl_value']
        # Create an empty DataFrame with the specified columns
        DTC_c1 = pd.DataFrame(columns=columns) 
        DTC_raw = pd.DataFrame()
        # Drop duplicated and null values of uniprot
        input_protein = input_protein.dropna(subset=['uniprot'])
        input_protein = input_protein.drop_duplicates(subset='uniprot').reset_index(drop=True)  
        # Check if there are any input proteins remaining      
        if len(input_protein) > 0:
            input_protein_id = input_protein['uniprot'][0]

            # Filter DTC by protein of interest using uniprot ID
            DTC_raw = self.data_manager.retrieve_raw_data('target_id', input_protein_id)
            # Check if there are any compounds remaining
            if len(DTC_raw) > 0:

                DTC_filt = self._filter_database(DTC_raw)

                if len(DTC_filt) > 0:

                    pc = PubChemServer()
                    DTC_c = pd.DataFrame()
                    
                    if 'CID' in DTC_filt.columns:
                        # Retrieve compounds using cids
                        compounds = self._pubchem_search_cid(DTC_filt, columns, pc)

                        if len(compounds) > 0:
                            # Add additional information to interacting compounds and standardize its values       
                            DTC_c = pd.merge(compounds, DTC_filt, on='CID', how='left')
                    else:
                        # Retrieve compounds using compound id (chembl ids) and inchis if formers miss
                        compounds = self._pubchem_search_chembl(DTC_filt, 'compound_id', columns, pc)

                        if len(compounds) > 0:
                            # Add additional information to interacting compounds and standardize its values       
                            DTC_c = pd.merge(compounds, DTC_filt, left_on='id', right_on='compound_id', how='left')
                    
                    # Check if there are any compounds remaining
                    if len(DTC_c) > 0:
                        # Standardize database
                        DTC_act = self._standardize_database(DTC_c, dtc_mutated, pChEMBL_thres)

                        if len(DTC_act) > 0:
                            # Extend compounds information
                            DTC_c1 = DTC_act[columns[:-2]+['CID', 'compound_id', 'standard_type', 'standard_value', 'standard_units', 
                                                'pchembl_value', 'activity_comment']].drop_duplicates()
                            # Add additional information
                            DTC_c1['notes'] = np.nan
                            DTC_c1['datasource'] = 'DTC'
                            statement = 'completed'
                        else:
                            statement = 'Standardization reduced interactions to 0'
                    else:
                        statement = 'Compounds not found using Pubchem and ChEMBL'
                else:
                    statement = 'First filter reduced interactions to 0'
            else:
                statement = 'No interaction data'
        else:
            statement = 'Input protein doesn\'t contain uniprot id'                         
        
        return DTC_c1, statement, DTC_raw


    def _standard_converter(self, DTC_act: pd.DataFrame):
        """
        Converts all types from DTC database to those needed for pCHEMBL by storing the respective
        standardized value. Also converts all units to nM as it is required to compute pCHEMBL.

        Parameters
        ----------
        DTC_act : DataFrame
            DTC dataframe with compounds whose types and units need to be standardized

        Returns
        -------
        DataFrame
            DTC dataframe of compounds with standardized types and unit
        """    

        std_types = ['IC50','KI','EC50','KD','Kd','ED50','XC50','AC50','Ki','ED50','CC50','AVERAGEIC50',
                'ACTIVITYEC50','SC50','AD50','DC50','XI50','GC50','GI50']
        log_types = ['LOGEC50','LOGIC50','LOGKD','LOGKI','logIC50','LOGIC50','LOGEC50','LOGKD']
        nlog_types = ['-LOGIC50','PIC50','-LOGKD','PEC50','PIC50(CALC)','-LOGED50','PED50','PXC50']
        replog_types = ['LOG1/IC50','LOG1/KD','LOG(1/IC50)']
        
        # Converting standard types
        DTC_act['standardized_val'] = None
        D1 = DTC_act.loc[DTC_act['standard_type'].isin(std_types)]
        for h in D1.index:
            DTC_act.loc[h,'standardized_val'] = DTC_act['standard_value'][h]
        D1 = DTC_act.loc[DTC_act['standard_type'].isin(nlog_types)]
        for h in D1.index:
            DTC_act.loc[h,'standardized_val'] = 10**(-DTC_act['standard_value'][h])
        D1 = DTC_act.loc[DTC_act['standard_type'].isin(log_types)]
        for h in D1.index:
            DTC_act.loc[h,'standardized_val'] = 10**(DTC_act['standard_value'][h])
        D1 = DTC_act.loc[DTC_act['standard_type']=='LOG(10^6/IC50)']
        for h in D1.index:
            DTC_act.loc[h,'standardized_val'] = 10**(6-(DTC_act['standard_value'][h]))
        D1 = DTC_act.loc[DTC_act['standard_type'].isin(replog_types)] 
        for h in D1.index:
            DTC_act.loc[h,'standardized_val'] = 10**(-DTC_act['standard_value'][h])
        D1 = DTC_act.loc[DTC_act['standard_type']=='1/KI'] 
        for h in D1.index:
            DTC_act.loc[h,'standardized_val'] = (1/DTC_act['standard_value'][h])
        
        # Converting all standard units to nM
        types = ['NM','NMOL/L']
        D1 = DTC_act.loc[DTC_act['standard_units'].isin(types)]
        for h in D1.index:
            DTC_act.loc[h,'standardized_val']=DTC_act['standardized_val'][h]
        types = ['M']
        D1 = DTC_act.loc[DTC_act['standard_units'].isin(types)]
        for h in D1.index:
            DTC_act.loc[h,'standardized_val'] = DTC_act['standardized_val'][h]*(1e9)
        types = ['/NM']
        D1 = DTC_act.loc[DTC_act['standard_units'].isin(types)]
        for h in D1.index:
            DTC_act.loc[h,'standardized_val'] = 1/DTC_act['standardized_val'][h]
        types = ['MM']
        D1=DTC_act.loc[DTC_act['standard_units'].isin(types)]
        for h in D1.index:
            DTC_act.loc[h,'standardized_val']=DTC_act['standardized_val'][h]*(1e9/1e3)
        types=['10^5/M']
        D1=DTC_act.loc[DTC_act['standard_units'].isin(types)]
        for h in D1.index:
            DTC_act.loc[h,'standardized_val']=((1/(DTC_act['standardized_val'][h]*(1e5)))*1e9)
        types=['NM G-1']
        D1=DTC_act.loc[DTC_act['standard_units'].isin(types)]
        for h in D1.index:
            DTC_act.loc[h,'standardized_val']=DTC_act['standardized_val'][h]*(1e3)
        types=['M-1']
        D1=DTC_act.loc[DTC_act['standard_units'].isin(types)]
        for h in D1.index:
            DTC_act.loc[h,'standardized_val']=((1/DTC_act['standardized_val'][h])*1e9)
        types=['/UM']
        D1=DTC_act.loc[DTC_act['standard_units'].isin(types)]
        for h in D1.index:
            DTC_act.loc[h,'standardized_val']=(1/(DTC_act['standardized_val'][h]))*(1e9/1e6)
        types=['PM']
        D1=DTC_act.loc[DTC_act['standard_units'].isin(types)]
        for h in D1.index:
            DTC_act.loc[h,'standardized_val']=DTC_act['standardized_val'][h]*(1e9/1e12)
        types=['MOL/G']
        D1=DTC_act.loc[DTC_act['standard_units'].isin(types)]
        for h in D1.index:
            DTC_act.loc[h,'standardized_val']=DTC_act['standardized_val'][h]*(1e3/1)*(1e9/1)
        types=['NMOL/MG']
        D1=DTC_act.loc[DTC_act['standard_units'].isin(types)]
        for h in D1.index:
            DTC_act.loc[h,'standardized_val']=DTC_act['standardized_val'][h]*(1e3/1e-3)
        types=['UM L-1','UMOL.KG-1']  
        D1=DTC_act.loc[DTC_act['standard_units'].isin(types)]
        for h in D1.index:
            DTC_act.loc[h,'standardized_val']=DTC_act['standardized_val'][h]*(1e9/1e6)
        
        types=['UG.ML-1','UG ML-1','UG/ML']
        D1=DTC_act.loc[DTC_act['standard_units'].isin(types)]
        for h in D1.index:
            if D1['molecular_weight'][h]=='':
                DTC_act.loc[h,'standardized_val']=np.nan
            else:
                DTC_act.loc[h,'standardized_val']=DTC_act['standardized_val'][h]*(1e3/1)*(1/1e6)*(1/float(D1['molecular_weight'][h]))*(1e9/1)
        types=['MG/ML']
        D1=DTC_act.loc[DTC_act['standard_units'].isin(types)]
        for h in D1.index:
            if D1['molecular_weight'][h]=='':
                DTC_act.loc[h,'standardized_val']=np.nan
            else:
                DTC_act.loc[h,'standardized_val']=DTC_act['standardized_val'][h]*(1e3/1)*(1/1e3)*(1/float(D1['molecular_weight'][h]))*(1e9/1)
        types=['NG ML-1']
        D1=DTC_act.loc[DTC_act['standard_units'].isin(types)]
        for h in D1.index:
            if D1['molecular_weight'][h]=='':
                DTC_act.loc[h,'standardized_val']=np.nan
            else:
                DTC_act.loc[h,'standardized_val']=DTC_act['standardized_val'][h]*(1e3/1)*(1/1e9)*(1/float(D1['molecular_weight'][h]))*(1e9/1)
        types=['MG.KG-1','MG KG-1']
        D1=DTC_act.loc[DTC_act['standard_units'].isin(types)]
        for h in D1.index:
            if D1['molecular_weight'][h]=='':
                DTC_act.loc[h,'standardized_val']=np.nan
            else:
                DTC_act.loc[h,'standardized_val']=DTC_act['standardized_val'][h]*(1/1e3)*(1/float(D1['molecular_weight'][h]))*(1e9/1)
        types=['UG KG-1']
        D1=DTC_act.loc[DTC_act['standard_units'].isin(types)]
        for h in D1.index:
            if D1['molecular_weight'][h]=='':
                DTC_act.loc[h,'standardized_val']=np.nan
            else:
                DTC_act.loc[h,'standardized_val']=DTC_act['standardized_val'][h]*(1/1e6)*(1/float(D1['molecular_weight'][h]))*(1e9/1)
        types=['P.P.M.','PPM']
        D1=DTC_act.loc[DTC_act['standard_units'].isin(types)]
        for h in D1.index:
            if D1['molecular_weight'][h]=='':
                DTC_act.loc[h,'standardized_val']=np.nan
            else:
                DTC_act.loc[h,'standardized_val']=DTC_act['standardized_val'][h]*(1/float(D1['molecular_weight'][h]))*(1/1e3)*(1e9/1)
        types=["10'13NM","10'8NM","10'7NM"]
        D1=DTC_act.loc[DTC_act['standard_units'].isin(types)]
        for h in D1.index:
            E=DTC_act['standard_units'][h][DTC_act['standard_units'][h].find("'")+1:DTC_act['standard_units'][h].find("N")]
            DTC_act.loc[h,'standardized_val']=DTC_act['standardized_val'][h]*(10**int(E))
        types=["10'-10M","10'-8M","10'-5M","10'-7M","10'-9M"]
        D1=DTC_act.loc[DTC_act['standard_units'].isin(types)]
        for h in D1.index:
            E=DTC_act['standard_units'][h][DTC_act['standard_units'][h].find("'")+1:DTC_act['standard_units'][h].find("M")]
            DTC_act.loc[h,'standardized_val']=DTC_act['standardized_val'][h]*(10**int(E))*(1e9)
        types=["10'-4UMOL/L"]
        D1=DTC_act.loc[DTC_act['standard_units'].isin(types)]
        for h in D1.index:
            E=DTC_act['standard_units'][h][DTC_act['standard_units'][h].find("'")+1:DTC_act['standard_units'][h].find("U")]
            DTC_act.loc[h,'standardized_val']=DTC_act['standardized_val'][h]*(10**int(E))*(1e9/1e6)
        types=["10'-9L/MOL","10'-10L/MOL"]
        D1=DTC_act.loc[DTC_act['standard_units'].isin(types)]
        for h in D1.index:
            E=DTC_act['standard_units'][h][DTC_act['standard_units'][h].find("'")+1:DTC_act['standard_units'][h].find("L")]
            DTC_act.loc[h,'standardized_val']=(1/(DTC_act['standardized_val'][h]*(10**int(E))))*(1e9)
        
        return DTC_act



from .DataManager import DataManager

class APIManager(DataManager):

    def retrieve_raw_data(self, filter_column, filter_value, **kwargs):
        
        # Call API function specified by filter_column
        raw_data = self.data[filter_column](filter_value)
        
        return raw_data


from .APIManager import APIManager
from .DataManager import DataManager
from .LocalManager import LocalManager
from .SQLManager import SQLManager

__all__ = ['APIManager', 'DataManager', 'LocalManager', 'SQLManager']


import pandas as pd
import numpy as np
import time
from collections import defaultdict
from ..servers.BiomartServer import BiomartServer
from ..servers.PubchemServer import PubChemServer
from ..data_manager import *
from .Database import Database

class DB(Database):

    def __init__(self, connection=None, database=None):
        # if not connection and not database:
        #     raise ValueError('Either SQL connection or database should be not None')
        if database is not None:
            self.data_manager = LocalManager(database)
        else:
            self.data_manager = SQLManager(connection, 'DB')


    def interactions(self, input_comp: pd.DataFrame):
        """
        Retrieves proteins from DrugBank database interacting with compound passed as input.

        Steps
        -----
        - Finds matches for all synonyms of the compound passed as input.
        - Finds all proteins (targets, enzymes, carriers, transporters) interacting with input compound. \\
        Constraints:
            - Only Homo Sapiens interactions \\
        - Uses Biomart to obtain proteins info (modified with data from original DB database) to return,
        searching with GeneID.
        
        Parameters
        ----------
        DB_data : DataFrame
            Dataframe containing all DrugBank database info
            
        input_comp : DataFrame
            Dataframe of input compounds from which interacting proteins are found

        Returns
        -------
        DataFrame
            Dataframe of interacting proteins, containing the following values:
            - entrez
            - gene_type
            - hgnc_symbol
            - description
            - datasource (DrugBank)
            - pchembl_value (nan)

        String
            A statement string describing the outcome of the database search

        DataFrame
            Raw Dataframe containing all DrugBank info about the input compound
        """

        columns = ['entrez','gene_type','hgnc_symbol','description','pchembl_value','datasource']
        # Create an empty DataFrame with the specified columns 
        DB_act = pd.DataFrame(columns=columns) 
        # Create a drugbank activity dataframe 
        DB_raw = pd.DataFrame(columns=columns) 

        input_comp = input_comp.dropna(subset=['cid'])
        if len(input_comp) > 0:

            input_comp_id = input_comp['cid'][0]
            DB_raw = self.data_manager.retrieve_raw_data('CID', input_comp_id)

            # Try searching with synonyms instead
            if len(DB_raw) == 0:
                input_comp = input_comp.dropna(subset=['synonyms'])
                # Extract all synonyms into a single list
                all_synonyms = list(input_comp['synonyms'].explode().dropna().str.lower().unique())
                # Filter DB_data based on all unique synonyms
                DB_raw = self.data_manager.retrieve_raw_data('name', all_synonyms)

            # Check if at least one match has been found
            if len(DB_raw) > 0: 
 
                # Remove null values for drugbank-id and HGNC
                DB_act = DB_raw.dropna(subset=['drugbank-id', 'HGNC'])
                # Filter to Human only interactions
                DB_act = DB_act[DB_act['organism'] == 'Humans'].reset_index(drop=True) 

                # Check if at least one interaction has been found
                if len(DB_act) > 0:
                    # Unify gene identifiers by Converting DrugBank with Biomart
                    ensembl = BiomartServer()
                    # Search by hgnc id match to biomart
                    input_type='hgnc_id' 
                    names=['hgnc_id','entrez','gene_type','hgnc_symbol','description']
                    attributes = ['hgnc_id','entrezgene_id','gene_biotype','hgnc_symbol','description']
                    
                    db_targets = pd.DataFrame()
                    
                    input_genes = list(DB_act['HGNC'])
                    
                    db_targets = ensembl.subset_search(input_type, input_genes, attributes, names)
                        
                    # For each compound, assign specific biomart column values to the ones from the original DrugBank database
                    for index, row in DB_act.iterrows():
                        S1 = db_targets.loc[db_targets['hgnc_id']==row['HGNC']]
                        if len(S1) > 0:
                            DB_act.loc[index,'entrez'] = S1['entrez'].iloc[0]
                            DB_act.loc[index,'gene_type'] = S1['gene_type'].iloc[0]
                            DB_act.loc[index,'hgnc_symbol'] = S1['hgnc_symbol'].iloc[0]
                            DB_act.loc[index,'description'] = S1['description'].iloc[0]
                        else:
                            DB_act.loc[index, 'entrez'] = None
                            DB_act.loc[index, 'gene_type'] = None
                            DB_act.loc[index, 'hgnc_symbol'] = None
                            DB_act.loc[index, 'description'] = None  

                        DB_act['datasource'] = 'DrugBank'
                        DB_act['pchembl_value'] = np.nan
                        statement = 'completed'
                else:
                    statement = 'Filter reduced interactions to 0'
            else:
                statement = 'No interaction data'
        else:
            statement= 'Compound doesn\'t have any cids'        
        return DB_act, statement, DB_raw


    def compounds(self, input_protein: pd.DataFrame):
        """
        Retrieves compounds from db database interacting with proteins passed as input.

        Steps
        -----
        - Filters stitch database to find compounds interacting with input proteins: \\
        Constraints:
            - organism = Humans
            - compound must have inchikey
        - Uses Pubchempy to obtain compound info (modified with data from original bdb database) to return, 
        searching with inchikey.
        
        Parameters
        ----------
        input_protein : DataFrame
            Dataframe of input proteins from which interacting compound are found
        DB_data: DataFrame
            Dataframe containing all drugbank database info

        Returns
        -------
        DataFrame
            Dataframe of interacting compounds, containing the following values: \\
            inchi, inchikey, isomeric_smiles, iupac_name, datasource (db), pchembl_value (nan), notes (nan)
        """

        columns = ['inchi','inchikey','isomeric_smiles','iupac_name','datasource','pchembl_value']
        # Create an empty DataFrame with the specified columns
        db_c1 = pd.DataFrame(columns=columns)
        DB_raw = pd.DataFrame()
        # Drop duplicated and null values of HGNC id
        input_protein = input_protein.dropna(subset=['hgnc_id'])
        input_protein = input_protein.drop_duplicates(subset='hgnc_id').reset_index(drop=True)

        # Check if there are any input proteins remaining
        if len(input_protein) > 0:
            # compound_data = defaultdict(list)
            # Search compounds interacting with input gene
            input_protein_id = input_protein['hgnc_id'][0]

            DB_raw = self.data_manager.retrieve_raw_data('HGNC', input_protein_id)
            
            # Check if at least one interaction has been found
            if len(DB_raw) > 0:

                # Filter to Human only interactions
                DB_act = DB_raw[DB_raw['organism'] == 'Humans'].copy().reset_index(drop=True)                    
                                 
                if len(DB_act) > 0:

                    pc = PubChemServer()

                    if 'CID' in DB_act:
                        # Retrieve compounds using cids
                        compounds = self._pubchem_search_cid(DB_act, columns, pc)
                        
                        if len(compounds) > 0:
                            db_c1 = pd.concat([db_c1, compounds])
                            db_info = DB_act[['CID', 'drugbank-id', 'name', 'HGNC']].\
                                rename(columns={'drugbank-id': 'compound_id', 
                                                'name': 'compound_name', 'HGNC': 'hgnc_id'}).drop_duplicates()
                            
                            db_c1 = pd.merge(db_c1, db_info, on='CID')
                    else:
                        # Remove duplicates and null values for inchikey
                        DB_act = DB_act.dropna(subset=['InChIKey']).drop_duplicates(subset=['InChIKey'])\
                            .reset_index(drop=True)    
                        compounds = pd.DataFrame()
                        selected_columns = pc.get_columns(columns[:-2])
                        # Perform search using Inchi key
                        for inchi in DB_act['InChIKey']:
                            try:
                                # Retrieve compound using pubchempy
                                comp = (pc.get_compounds(inchi, selected_columns, namespace='inchikey'))
                                compounds = pd.concat([compounds, comp])
                            except:
                                None
                            time.sleep(0.5)

                        if len(compounds) > 0:
                            db_c1 = pd.concat([db_c1, compounds.rename(columns=pc.properties)])
                            db_info = DB_act[['InChIKey', 'drugbank-id', 'name', 'HGNC']].\
                                    rename(columns={'InChIKey': 'inchikey', 'drugbank-id': 'compound_id', 
                                                    'name': 'compound_name', 'HGNC': 'hgnc_id'}).drop_duplicates()
                            # Add additional values from activity dataframe
                            db_c1 = pd.merge(db_c1, db_info, on='inchikey', how='left')

                    # Check if at least one compound has been found
                    if len(db_c1) > 0:
                        db_c1['datasource'] = 'DrugBank'
                        db_c1['pchembl_value'] = np.nan
                        db_c1['notes'] = np.nan
                        statement = 'completed'
                    else:
                        statement = 'Compounds not found using Pubchem'
                else:
                    statement = 'Filter reduced interactions to 0'
            else:
                statement = 'No interaction data'
        else:
            statement = 'Input protein doesn\'t contain HGNC'                  

        return db_c1, statement, DB_raw


    # def _find_compound_info(self, DB_act):
    #     """
    #     Extract compound info (inchikey) from DrugBank database and stores it into a new column.
    
    #     Parameters
    #     ----------
    #     DB_act : DataFrame
    #         DrugBank database

    #     Returns
    #     -------
    #     DataFrame
    #         Drugbank database with a column containing compound inchikey 
    #     """
    #     for index, row in DB_act.iterrows():
    #         if type(row['calculated-properties']) == dict:
    #             for prop in row['calculated-properties'].values():
    #                 try:
    #                     if prop['kind'][0][0] == 'InChIKey':
    #                         DB_act.loc[index, 'InChIKey'] = prop['value'][0][0]
    #                         break
    #                 except:
    #                     None
    #     return DB_act


    # def preprocess_db(self, DB_data):

    #     DB_data['name'] = DB_data['name'].apply(lambda x: x[0][0].lower() if x else None)
    #     DB_data['drugbank-id'] = DB_data['drugbank-id'].apply(lambda x: x[0][0] if x else None)

    #     # Add inchikey column
    #     DB_data['InChIKey'] = None
    #     # Extract inchikey
    #     DB_data = self._find_compound_info(DB_data)

    #     protein_data = defaultdict(list)
    #     for _, row in DB_data.iterrows():
    #         hgncs = []
    #         organisms = []
    #         protein_names = []
    #         protein_types = []
    #         for prot_type in ['targets', 'carriers', 'enzymes', 'transporters']:
    #             if row[prot_type]:
    #                 for prot in row[prot_type].values():
    #                     try:
    #                         HGNC = prot['polypeptide']['external-identifiers']['external-identifier']['identifier'][0][0]
    #                         try:
    #                             organism = prot['organism'][0][0]
    #                         except:
    #                             organism = None
    #                         protein_name = prot['name'][0][0]
    #                     except (KeyError, TypeError):
    #                         HGNC = None
    #                         organism = None
    #                         protein_name = None
    #                     hgncs.append(HGNC)
    #                     organisms.append(organism)
    #                     protein_names.append(protein_name)
    #                     protein_types.append(prot_type)
    #         protein_data['HGNCs'].append(hgncs)
    #         protein_data['organisms'].append(organism)
    #         protein_data['protein_names'].append(protein_names)
    #         protein_data['types'].append(protein_types)

    #     new_data = pd.DataFrame({'HGNC': protein_data['HGNCs'], 
    #                                     'organism': protein_data['organisms'], 
    #                                     'protein_name': protein_data['protein_names'],
    #                                     'protein_type': protein_data['types']})

    #     DB_data = pd.concat([DB_data, new_data], axis=1)
    #     DB_data = DB_data.explode(['HGNC', 'protein_name', 'protein_type']).reset_index(drop=True)
    #     DB_data = DB_data[['drugbank-id', 'name', 'InChIKey', 'HGNC', 'organism', 'protein_name', 'protein_type']]
    #     return DB_data



import time
from chembl_webresource_client.new_client import new_client
import pandas as pd

class ChEMBLServer(object):
    
    @staticmethod
    def identify_chembl_ids(compound: pd.DataFrame, pubchem_ids=True):
        chembl_total = []
        if pubchem_ids:
            # Search for all the chembl ids for the compound
            for _, row in compound.iterrows():
                syns = row['synonyms']
                if syns:
                    chembl_total.extend([syn for syn in syns if (syn and 'CHEMBL' in syn)])
        else:
            # Collect a list of ChEMBL ids from the PubChem names
            chemblEx=[] 
            chemblSyn=[] 
            # Search for all the chembl ids of synonyms for the compound
            for _, row in compound.iterrows():
                syns = row['synonyms']
                for syn in syns:
                    try:
                        molecule = new_client.molecule # This uses the chembl webresource client to find chembl ids
                        molS = molecule.filter(molecule_synonyms__molecule_synonym__iexact=syn)\
                                                .only(['molecule_chembl_id'])
                        molE = molecule.filter(pref_name__iexact=syn).only(['molecule_chembl_id'])
                        # Add all Exact hits
                        if len(molE) > 0:
                            chemblEx.append(molE[0]['molecule_chembl_id']) 

                        # Add all Synonym hits
                        if len(molS) > 0:
                            chemblSyn.append(molS[0]['molecule_chembl_id'])
                    except:
                        # API not working, wait 5 seconds
                        time.sleep(5)
            chembl_total = chemblEx + chemblSyn
            
        # If search is with ChEMBL or PubChem search was not successful 
        if not pubchem_ids or (chembl_total == [] and pubchem_ids):  
            chemblKey=[] 
            chemblSmi=[]  
            # Find the inchikey mapping to ChEMBL id
            inchikey = compound['inchikey'][0]
            if pd.notna(inchikey):
                try:
                    molecule = new_client.molecule
                    mols = molecule.filter(molecule_structures__standard_inchi_key=inchikey).only(['molecule_chembl_id'])
                    # Add all inchikey hits
                    if len(mols) > 0:
                        chemblKey.append(mols[0]['molecule_chembl_id'])
                except:
                    time.sleep(5)

            # Find the SMILES mapping to ChEMBL id
            smiles = compound['isomeric_smiles'][0]
            if pd.notna(smiles):
                try:
                    molecule = new_client.molecule
                    # Add all smiles hits
                    mols=molecule.filter(molecule_structures__canonical_smiles__connectivity=smiles)\
                                            .only(['molecule_chembl_id'])
                    if len(mols) > 0:
                        chemblSmi.append(mols[0]['molecule_chembl_id'])
                except:
                    time.sleep(5)

            # Merge all hits 
            chembl_total.extend(chemblKey + chemblSmi)

        # Obtain a single list of unique ids
        chembl_total = list(set(chembl_total))
        if len(chembl_total) > 0:
            return chembl_total
        else:
            return [None]

    @staticmethod
    def molecule_search(ids, inchi, chembl_id):

        attempts = 0
        max_attempts = 10
        while True:
            try:
                molecule = new_client.molecule
                molecule.set_format('json')
                record_via_client = molecule.filter(molecule_chembl_id__in=ids).only(['molecule_chembl_id','molecule_structures'])
                for record in record_via_client:
                    if record['molecule_structures']:
                        inchi.append(record['molecule_structures']['standard_inchi'])#
                        chembl_id.append(record['molecule_chembl_id'])
                break
            except: 
                # Avoid the chembl API error, and give up after 10 attempts
                time.sleep(5)
                attempts += 1
                if attempts >= max_attempts:
                    print(f"attempts ({max_attempts}) reached. Chembl API does not work.")
                    break


from abc import ABC, abstractmethod

class DataManager(ABC):

    def __init__(self, data_source, db_name=None):
        self.data = data_source
        self.database = db_name

    @abstractmethod
    def retrieve_raw_data(self, filter_column, filter_value, **kwargs):
        raise NotImplementedError



import time
import mysql.connector

def connect_to_mysql(config, attempts=3, delay=2):
    attempt = 0
    # Implement a reconnection routine
    while attempt < attempts:
        try:
            return mysql.connector.connect(**config)
        except (mysql.connector.Error, IOError) as err:
            if (attempts is attempt):
                # Attempts to reconnect failed; returning None
                print("Failed to connect, exiting without a connection: %s", err)
                return None
            print("Connection failed: %s. Retrying (%d/%d)...",
                err,
                attempt,
                attempts-1)
            # progressive reconnect delay
            time.sleep(delay ** attempt)
            attempt += 1
    return None


from .DataManager import DataManager
import pandas as pd
import mysql

class SQLManager(DataManager):

    def retrieve_raw_data(self, filter_column, filter_value, **kwargs):

        if self.data is None:
            raise ValueError('SQL connection not available')
        
        raw_data = pd.DataFrame()
        with self.data.cursor() as cursor:
            try:
                # SQL query to select data from the table where the specified column has a certain value
                query = f"SELECT * FROM {self.database} WHERE `{filter_column}` in (%s)"
                if type(filter_value) == list:
                    values = ",".join(filter_value)
                    cursor.execute(query, (values,))
                else:
                    cursor.execute(query, (filter_value,))

                # Fetch all the rows
                rows = cursor.fetchall()

                # Create a DataFrame from the result
                columns = [desc[0] for desc in cursor.description]
                raw_data = pd.DataFrame(rows, columns=columns)

            # Connection error
            except mysql.connector.Error as err:
                print(f'Error while fetching data from SQL server: {err}')
            
            return raw_data


import pandas as pd
import numpy as np
from chembl_webresource_client.new_client import new_client
from ..servers.BiomartServer import BiomartServer
from ..servers.PubchemServer import PubChemServer 
from ..servers.ChEMBLServer import ChEMBLServer as chembl
from .Database import Database
from ..data_manager import *
import re

class ChEMBL(Database):

    def __init__(self, connection=None, database=None):
        if database is not None:
            self.data_manager = LocalManager(database)
        elif connection is not None:
            self.data_manager = SQLManager(connection, 'CHEMBL')
        else:
            funcs = {
                'Molecule ChEMBL ID' : self._retrieve_compounds,
                'Target ChEMBL ID' : self._retrieve_proteins
            }
            self.data_manager = APIManager(funcs) 

        # Mapping for API and local database column names
        self.column_names = {
            'molecule_chembl_id': 'Molecule ChEMBL ID',
            'activity_comment': 'Comment',
            'data_validity_comment': 'Data Validity Comment',
            'target_tax_id': 'Target Organism',
            'src_id': 'Source ID',
            'pchembl_value': 'pChEMBL Value',
            'target_chembl_id': 'Target ChEMBL ID'
        }

    def _filter_database(self, chembl_raw: pd.DataFrame, pChEMBL_thres: float):

        invalid_activities = ['Not Determined', 'Not Active', 'Not Evaluated', 'inactive', 'inconclusive', 'undetermined', 'No data']

        chembl_act = chembl_raw.loc[
                    # Remove all non-human
                    (chembl_raw['Target Organism'].isin(['9606', 'Homo sapiens'])) &  
                    # Remove various activity comments
                    (~chembl_raw['Comment'].isin(invalid_activities)) &  
                    # Only use interactions extracted from scientific literature
                    (chembl_raw['Source ID'] == 1) &  
                    # Only use interactions with no data validity comments
                    (chembl_raw['Data Validity Comment'].isnull()) &  
                    # Only use interactions with pchembl value
                    (chembl_raw['pChEMBL Value'].notnull())  
                ].drop_duplicates(subset=['Comment', 'Molecule ChEMBL ID', 'pChEMBL Value']).copy()
        chembl_act['pChEMBL Value'] = chembl_act['pChEMBL Value'].apply(lambda x: re.sub(r'[^0-9.]', '', x) if type(x) is str else x)
        chembl_act['pChEMBL Value'] = chembl_act['pChEMBL Value'].apply(pd.to_numeric, errors='coerce') 
        # Filter interactions by pChEMBL value
        chembl_act = chembl_act.loc[chembl_act['pChEMBL Value'] > pChEMBL_thres].reset_index(drop=True) 

        return chembl_act

    def interactions(self, input_comp: pd.DataFrame, chembl_ids: list=[], pChEMBL_thres=0):
        """
        Retrieves proteins from chembl API interacting with compound passed as input.

        Steps
        -----
        - Finds chembl id of all synonyms for the compound passed as input \\
        Constraints:
            - Only Homo Sapiens interactions
            - Remove various activity comments (Not Determined, Not Active, Not Evaluated, inactive, inconclusive, undetermined, No data)
            - Only interactions extracted from scientific literature
            - Only interactions with no data validity comments
            - Only interactions with pchembl value
        - Identifies target proteins using chembl API.
        - Uses Biomart to obtain proteins info (modified with data from first API search) to return,
        searching with chembl id.
        
        Parameters
        ----------
        input_comp : DataFrame
            Dataframe of input compounds from which interacting proteins are found
        chembl_ids : list
            Empty list that will contain all ChEMBL ids found for the input compound
        pChEMBL_thres : float
            minimum pChEMBL value necessary for interaction to be considered valid

        Returns
        -------
        DataFrame
            Dataframe of interacting proteins, containing the following values:
            - entrez 
            - gene_type 
            - hgnc_symbol 
            - description 
            - datasource (ChEMBL) 
            - pchembl_value

        String
            A statement string describing the outcome of the database search
        DataFrame
            Raw Dataframe containing all ChEMBL info about the input compound
        """
        
        columns = ['entrez','gene_type','hgnc_symbol','description','datasource', 'pchembl_value']
        # Create an empty DataFrame with the specified columns
        chembl_act = pd.DataFrame(columns=columns)
        chembl_raw = pd.DataFrame()
        # Find chembl ids from input compound
        chembl_ids.extend(chembl.identify_chembl_ids(input_comp))
        # Check if at least one id has been found
        if len(chembl_ids) > 0 and None not in chembl_ids:
 
            chembl_raw = self.data_manager.retrieve_raw_data('Molecule ChEMBL ID', chembl_ids)
            try:
                chembl_raw.reset_index()
            except:
                chembl_raw = pd.DataFrame()
            
            if len(chembl_raw) > 0:
                # Filter ChEMBL for high quality interactions. This follows: Bosc, N. et al. J. cheminformatics 11, 1–16 (2019)
                chembl_act = self._filter_database(chembl_raw, pChEMBL_thres)

                if len(chembl_act) > 0:
                    # Retrieve target type if not already available
                    if 'Target Type' not in chembl_act.columns:
                        chembl_act = self._retrieve_target_type(chembl_act)

                    # Keep only relevant columns
                    chembl_act = chembl_act[['Molecule ChEMBL ID', 'Comment', 'Data Validity Comment', 'Target Organism',
                                             'Source ID', 'pChEMBL Value', 'Target ChEMBL ID', 'Target Type']]
                    # Only use int. w/proteins, removes cell lines and RNA
                    chembl_act = chembl_act.loc[chembl_act['Target Type']=='SINGLE PROTEIN'] 
                    chembl_act = chembl_act.rename(columns={'pChEMBL Value': 'pchembl_value'}).reset_index(drop=True)
                    
                    if len(chembl_act) > 0:
                        # Unify gene identifiers by Converting ChEMBL with biomart
                        ensembl = BiomartServer()
                        # ChEMBL uses chembl ids
                        input_type='chembl' 
                        attributes = ['chembl', 'entrezgene_id', 'gene_biotype', 'hgnc_symbol', 'description']
                        names = ['chembl','entrez','gene_type','hgnc_symbol','description']
                        chembl_targets=pd.DataFrame(columns=names)
                        
                        input_genes=list(chembl_act['Target ChEMBL ID'])
                        
                        chembl_targets=ensembl.subset_search(input_type, input_genes, attributes, names)

                        # For each compound, assign specific biomart column values to the ones from the original chembl database
                        for index, row in chembl_act.iterrows():
                            S1 = chembl_targets.loc[chembl_targets['chembl']==row['Target ChEMBL ID']]
                            if len(S1) > 0:
                                chembl_act.loc[index, 'entrez'] = S1['entrez'].iloc[0]
                                chembl_act.loc[index, 'gene_type'] = S1['gene_type'].iloc[0]
                                chembl_act.loc[index, 'hgnc_symbol'] = S1['hgnc_symbol'].iloc[0]
                                chembl_act.loc[index, 'description'] = S1['description'].iloc[0]
                            else:
                                chembl_act.loc[index, 'entrez'] = None
                                chembl_act.loc[index, 'gene_type'] = None
                                chembl_act.loc[index, 'hgnc_symbol'] = None
                                chembl_act.loc[index, 'description'] = None  
                                Des='Failed to convert the gene ID'
                        chembl_act['datasource'] = 'ChEMBL'
                        statement = 'completed'
                    else:
                        statement = 'Filter reduced interactions to 0'
                else:
                    statement = 'Filter reduced interactions to 0'
            else:
                statement = 'No interaction data'
        else:
            statement = 'No ChEMBL ids found'

        return chembl_act, statement, chembl_raw
    
    def compounds(self, input_protein: pd.DataFrame, pChEMBL_thres: float=0):
        """
        Retrieves compounds from chEMBL database interacting with proteins passed as input.

        Steps
        -----
        - Makes an API call to chEMBL API to retrieve compounds interacting with input protein \\
        Constraints:
            - Only Homo Sapiens interactions
            - Remove various activity comments (Not Determined, Not Active, Not Evaluated, inactive, inconclusive, undetermined, No data)
            - Only interactions extracted from scientific literature
            - Only interactions with no data validity comments
            - Only interactions with pchembl value
        - Uses Pubchempy to retrieve the compound info (with additional data from chembl) to return, 
        searching with chEMBL ID or with inchi (obtained from chembl API) if the first search is unsuccessful.
        
        Parameters
        ----------
        input_protein : DataFrame
            Dataframe of input proteins from which interacting compound are found
        pChEMBL_thres : float
            minimum pChEMBL value necessary for interaction to be considered valid

        Returns
        -------
        DataFrame
            Dataframe of interacting compounds, containing the following values:
            - inchi
            - inchikey
            - isomeric smiles
            - iupac name
            - datasource (ChEMBL)
            - pchembl value
            - notes (NaN)
        """

        columns = ['inchi','inchikey','isomeric_smiles','iupac_name','datasource','pchembl_value']
        # Create an empty DataFrame with the specified columns
        chembl_c1 = pd.DataFrame(columns=columns) 
        chembl_raw = pd.DataFrame()
        # Drop duplicated and null value of chembl_id, to make sure the first line of the dataset has the chembl ID
        input_protein = input_protein.dropna(subset=['chembl'])
        input_protein = input_protein.drop_duplicates(subset='chembl').reset_index(drop=True)        
        # Check if there are any input proteins remaining
        if len(input_protein) > 0:
            input_protein_id = input_protein['chembl'].iloc[0]

            chembl_raw = self.data_manager.retrieve_raw_data('Target ChEMBL ID', input_protein_id)
            # Check if the API returned any interacting compound
            if len(chembl_raw) > 0:
                # Filter ChEMBL for high quality interactions.
                chembl_act = self._filter_database(chembl_raw, pChEMBL_thres)

                if len(chembl_act) > 0:

                    pc = PubChemServer()

                    if 'CID' in chembl_act.columns:
                        # Retrieve compounds using cids
                        compounds = self._pubchem_search_cid(chembl_act, columns, pc)
                        
                        if len(compounds) > 0:
                            chembl_info = chembl_act[['CID', 'Molecule ChEMBL ID', 'pChEMBL Value', 'Target ChEMBL ID']].\
                                            rename(columns={'Molecule ChEMBL ID': 'id', 'pChEMBL Value': 'pchembl_value', 
                                                            'Target ChEMBL ID': 'target_chembl_id'}).drop_duplicates()
                            chembl_c1 = pd.merge(compounds, chembl_info, on='CID', how='left')
                        
                    else:
                        # Retrieve compounds using chembl ids and inchis if formers miss
                        compounds = self._pubchem_search_chembl(chembl_act, 'Molecule ChEMBL ID', columns, pc)

                        if len(compounds) > 0:
                            # Update pchembl value with that of filtered compounds
                            chembl_info = chembl_act[['Molecule ChEMBL ID', 'pChEMBL Value', 'Target ChEMBL ID']].\
                                            rename(columns={'Molecule ChEMBL ID': 'id', 'pChEMBL Value': 'pchembl_value', 
                                                            'Target ChEMBL ID': 'target_chembl_id'}).drop_duplicates()
                            chembl_c1 = pd.merge(compounds, chembl_info, on='id', how='left')

                    # Check if at least one compound has been found in total
                    if len(chembl_c1) > 0:
                        # Store into dataframe additional data
                        chembl_c1.loc[:, 'datasource'] = 'ChEMBL'
                        chembl_c1.loc[:, 'notes'] = np.nan
                        statement = 'completed'
                    else:
                        statement = 'Compounds not found using Pubchem and ChEMBL'
                else:
                    statement = 'Filter reduced interactions to 0'
            else:
                statement = 'No interaction data'
        else:
            statement = 'Input protein doesn\'t contain chembl id'

        return chembl_c1, statement, chembl_raw
    

    def _retrieve_compounds(self, chembl_ids):
            chembl_raw = pd.DataFrame(columns=['entrez','gene_type','hgnc_symbol','description','datasource', 'pchembl_value'])
            # API to Obtain all the targets for the list of chembl ids and their reported activities
            for id in chembl_ids: 
                # Use the chembl webresource client to find interactions
                activities = new_client.activity.filter(molecule_chembl_id=id)\
                .only(['molecule_chembl_id','activity_comment','data_validity_comment','target_tax_id',\
                       'src_id','pchembl_value','target_chembl_id'])
                if len(activities) != 0:
                    mol_act = pd.DataFrame(activities)
                    chembl_raw = pd.concat([chembl_raw, mol_act])
            
            chembl_raw = chembl_raw.rename(columns=self.column_names)

            return chembl_raw
    
    def _retrieve_target_type(self, chembl_act):
        # Filter ChEMBL based on target_type, looking to only use proteins
        target = new_client.target
        dat = []
        # Use the chembl webresource client to find target_type
        for _, row in chembl_act.iterrows(): 
            res = target.filter(target_chembl_id=row['Target ChEMBL ID']).only(['target_type'])
            dat.append(res[0]['target_type'])
        chembl_act['Target Type'] = dat

        return chembl_act

    def _retrieve_proteins(self, input_protein_id):
        # Use package of Chembl API to get the compound that have an activity with the protein
        activities = new_client.activity.filter(target_chembl_id=input_protein_id)\
                    .only(['molecule_chembl_id','activity_comment','data_validity_comment','target_tax_id',\
                            'src_id','pchembl_value','target_chembl_id'])
        # Create dataframe from API output
        chembl_raw = pd.DataFrame.from_dict(activities)
        chembl_raw = chembl_raw.rename(columns=self.column_names)

        return chembl_raw


# import pandas as pd
# import numpy as np

def generate_subsets(set_size: int, n: int):
    """
    Generate starting and ending subset indices, diving set_size into n equal subsets
    
    Parameters
    ----------
    set_size : int
        size of the set to divide
    n : int
        number of subset to generate

    Yields
    -------
    int
        index of the first element of current subset
    int
        index of the last element of current subset
    """   
    start = 0
    while start < set_size:
        end = min(start + n, set_size)
        yield start, end
        start = end

class Singleton(type):
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]
    



import numpy as np
import pandas as pd
from ..databases import *
from ..utils.identifiers import protein_identifiers
from .Pipeline import Pipeline

class Prot2Comp(Pipeline):

    def _update_args(self, pChEMBL_thres, stitch_stereo, dtc_mutated, dc_extra):
        self.database_args = {
            'pc': (pChEMBL_thres,),
            'chembl': (pChEMBL_thres,),
            'bdb': (pChEMBL_thres,),
            'stitch': (stitch_stereo,),
            'ctd': (),
            'dtc': (dtc_mutated, pChEMBL_thres,),
            'otp': (),
            'dc': (dc_extra, pChEMBL_thres,),
            'db': (),
        }      

    # Calls functions to collect data and merges all the data from the various sources together
    # The parameters are:
    #    - input_id - the protein id
    #    - pChEMBL_thresh - the minimum interaction pChEMBL value required to be added to the output file
    #    - stitch_stereo - to select whether to consider the specific compound stereochemistry or group all stereoisomers interactions from STITCH
    #    - dtc_mutated - to select whether also to consider interactions with mutated target proteins from DTC
    #    - dc_extra - to select whether to include possibly non-Homo sapiens interactions

    def prot_interactions(self, input_id, pChEMBL_thres=0, stitch_stereo=True, 
                          dtc_mutated=False, dc_extra=False):

        # Run interaction select with all databases selected
        prot_comp, state = self.prot_interactions_select(input_id, pChEMBL_thres=pChEMBL_thres, stitch_stereo=stitch_stereo, 
                                                         dtc_mutated=dtc_mutated, dc_extra=dc_extra)
        
        return prot_comp, state
            

    # Calls functions to collect data and merges all the data from the selected sources together
    # The parameters are:
    #    - input_id - the protein id
    #    - selected_dbs - underscore-separated string containing the names of the databases which are selected
    #    - pChEMBL_thresh - the minimum interaction pChEMBL value required to be added to the output file
    #    - stitch_stereo - to select whether to consider the specific compound stereochemistry or group all stereoisomers interactions from STITCH
    #    - dtc_mutated - to select whether also to consider interactions with mutated target proteins from DTC
    #    - dc_extra - to select whether to include possibly non-Homo sapiens interactions

    def prot_interactions_select(self, input_id, selected_dbs='pc_chembl_bdb_stitch_ctd_dtc_otp_dc_db', 
                                 pChEMBL_thres=0, stitch_stereo=True, dtc_mutated=False, dc_extra=False):

        self._update_args(pChEMBL_thres, stitch_stereo, dtc_mutated, dc_extra)

        prot_ids = protein_identifiers(input_id)

        main_columns=['inchi','inchikey','isomeric_smiles','iupac_name','pchembl_value','datasource']    

        prot_comp = pd.DataFrame(columns=main_columns)

        # Outputs the brief statement regarding the processing of each source 
        states=pd.DataFrame(columns=['prot'])
        states.loc[0, 'prot'] = input_id   
        
        if len(prot_ids) > 0:
            comp_all = pd.DataFrame(columns=main_columns)
            db_list = selected_dbs.split('_')
            for name, db in self.databases.items():      
                if name in db_list:
                    # Perform database search
                    db_comp, state, _ = db.compounds(prot_ids, *self.database_args.get(name, ()))
                    if len(db_comp) > 0:
                        # Keep only wanted columns
                        db_comp = db_comp.loc[:, main_columns]
                        # Concatenate the source
                        comp_all = pd.concat([comp_all, db_comp])
                    print(f'{name} done!')
                else:
                    state = f'Did not run {name}'
                # Add statement for this source
                states.loc[0, name] = state

            prot_comp = self._postprocess_databases(comp_all)
                        
            if len(prot_comp) > 0:
                # Add std compound ids to output
                std_ids = prot_ids[['entrez', 'gene_type', 'hgnc_symbol', 'description']]
                # Select first nonNa value from each column
                std_ids = std_ids.apply(lambda col: std_ids[col.name].dropna().iloc[0] if not std_ids[col.name].dropna().empty else None)
                # Assign to each row the standard ids
                prot_comp = prot_comp.assign(**std_ids)
                # Add input id to the results
                prot_comp['input_id'] = input_id

                # Reorder columns
                prot_comp = prot_comp[['input_id', 'entrez', 'gene_type', 'hgnc_symbol', 'description',
                                    'inchi', 'inchikey', 'isomeric_smiles', 'iupac_name', 'pchembl_count',
                                    'ave_pchembl', 'std_pchembl', 'src_count', 'pubchem', 'chembl', 'bindingdb', 'stitch',
                                    'ctd', 'dtc', 'otp', 'drugcentral', 'drugbank']]

        return prot_comp, states
            

    def _postprocess_databases(self, comp_all):
        comp_all = comp_all.dropna(subset=['inchi']).reset_index(drop=True)

        # Find the compound with same inchikey in 1st and 2nd block. 
        comp_all['12Key'] = comp_all['inchikey'].apply(lambda x: x[:-1]) 
        comp_all['3Key'] = comp_all['inchikey'].apply(lambda x: x[-1:])

        # Make a list of the unique targets for the protein
        comp_list = comp_all['12Key'].unique()
        tar_comp=pd.DataFrame(columns=['inchi','inchikey','isomeric_smiles','iupac_name',
                                    'pchembl_count','ave_pchembl','src_count'] + 
                                    [source.lower() for source in self.sources]
                                    +['notes'])

        for index, compound in enumerate(comp_list):

            comp = comp_all.loc[comp_all['12Key'] == compound]
            comp1 = comp.drop_duplicates()
            if len(comp) > 0:
                # Make sure the neutral one is in the final output
                is_N_present = comp1['3Key'].str.contains('N') 
                indexes_with_N = comp1.loc[is_N_present].index
                if len(indexes_with_N) > 0:
                    tar_comp.loc[index,'inchi'] = comp1['inchi'][indexes_with_N[0]]
                    tar_comp.loc[index,'inchikey'] = comp1['inchikey'][indexes_with_N[0]]
                    tar_comp.loc[index,'isomeric_smiles'] = comp1['isomeric_smiles'][indexes_with_N[0]]
                    tar_comp.loc[index,'iupac_name'] = comp1['iupac_name'][indexes_with_N[0]]
                    # tar_comp.loc[index,'pchembl_value'] = comp1['pchembl_value'][indexes_with_N[0]]
                else:
                    tar_comp.loc[index,'inchi'] = comp1['inchi'][comp.index[0]]
                    tar_comp.loc[index,'inchikey'] = comp1['inchikey'][comp.index[0]]
                    tar_comp.loc[index,'isomeric_smiles'] = comp1['isomeric_smiles'][comp.index[0]]
                    tar_comp.loc[index,'iupac_name'] = comp1['iupac_name'][comp.index[0]]
                    # tar_comp.loc[index,'pchembl_value'] = comp1['pchembl_value'][comp.index[0]]
            
            tar_comp = self._aggregate_pchembl(tar_comp, index, comp)

            tar_comp.loc[index, 'src_count'] = len(comp['datasource'].unique())
            # Sets a source binary matrix to the dataframe to easily filter output later
            for source in self.sources:
                tar_comp.loc[index, source.lower()] = 1 if source in comp['datasource'].unique() else 0

        return tar_comp
    



import pandas as pd
import numpy as np
import requests
import time
import json
from ..servers.BiomartServer import BiomartServer
from ..servers.ChEMBLServer import ChEMBLServer as chembl
from ..servers.PubchemServer import PubChemServer
from ..data_manager import *
from .Database import Database

class OTP(Database):

    def __init__(self, connection=None, database=None):
        if connection is not None or database is not None:
            raise ValueError('No SQL connection or database expected for OTP')
        funcs = {
            'bib_comps' : self._retrieve_bib_compounds,
            'met_comps' : self._retrieve_met_compounds,
            'proteins' : self._retrieve_proteins 
        }
        self.data_manager = APIManager(funcs)

    def interactions(self, input_comp: pd.DataFrame, chembl_ids: list, use_biblio: bool=False):
        """
        Retrieves proteins from OTP API interacting with compound passed as input.

        Steps
        -----
        - Retrieves Mechanism and Bibliography interactions, based on the use_biblio value, using the chembl
        ids passed as input.
        - Uses Biomart to obtain proteins info (modified with data from first API searchs) to return,
        searching with ensembl peptide id.
        
        Parameters
        ----------
        input_comp : DataFrame
            Dataframe of input compounds from which interacting proteins are found
        chembl_ids : list
            list of chembl ids for the input compounds
        use_biblio : bool
            value to determine which kind of interactions to retrieve:
            - False (default) - mechanism data only 
            - True - mechanism and bibliography \\ 
            False ensures real interactions, avoids keyword search matches within abstracts that can lead to false associations that are not real interactions
            
        Returns
        -------
        DataFrame
            Dataframe of interacting proteins, containing the following values: \\
            entrez, gene_type, hgnc_symbol, description, datasource (OTP), pchembl_value (nan)
        String
            A statement string describing the outcome of the database search
        DataFrame
            Raw Dataframe containing all OTP info about the input compound
        """
                
        columns = ['entrez','gene_type','hgnc_symbol','description','pchembl_value','datasource']
        # Create an empty DataFrame with the specified columns
        otp_act=pd.DataFrame(columns=columns)
        # Create an OTP activity DataFrame
        raw_columns = ['Chembl','PMID','mID','mLabel','mType','strength']
        otp_raw = pd.DataFrame(columns=raw_columns)
        
        # Check if chembl_ids has already been computed from the input compound, otherwise do so
        if not chembl_ids:
            chembl_ids.extend(chembl.identify_chembl_ids(input_comp))

        if len(chembl_ids) > 0 and None not in chembl_ids:
            # If True then takes both mechanisms and bibliography sections of OTP
            if use_biblio: 
                otp_bib = self.data_manager.retrieve_raw_data('bib_comps', chembl_ids)
            # Else only Mechanisms is collected from OTP, skipping bibliography
            otp_met = self.data_manager.retrieve_raw_data('met_comps', chembl_ids)
            
            # Merges bibliography and mechanism findings only if use_biblio is True
            if use_biblio:
                otp_raw = pd.concat([otp_bib,otp_met])
            else:
                otp_raw = otp_met

            if len(otp_raw) > 0:  
                # Filter interactions
                # Remove interactions without at least 1 source
                otp_filtered = otp_raw.loc[(otp_raw['PMID'].notnull()) &  
                                    # Removes interactions that are not gene interactions
                                    (otp_raw['mType']=='GP')].reset_index(drop=True) 
                # Note: OTP is human only, so no animal data here
                # Note: OTP does not collect activity data so we cannot calculate pChEMBL
                
                # Check if there is at least one interaction remaining 
                if len(otp_filtered) > 0:
                    otp_act = pd.concat([otp_act, otp_filtered])
                    # Unify gene identifiers by Converting OTP with biomart
                    ensembl = BiomartServer()
                    attributes = ['ensembl_gene_id', 'entrezgene_id', 'gene_biotype', 'hgnc_symbol', 'description']
                    names = ['ensembl_id','entrez','gene_type','hgnc_symbol','description']
                    # Search by ensembl peptide id match to biomart
                    input_type='ensembl_gene_id'
                    
                    otp_targets = pd.DataFrame(columns=names)
                    
                    input_genes = list(otp_act['mID'])
                    
                    otp_targets = ensembl.subset_search(input_type, input_genes, attributes, names)
                    
                    # For each compound, assign specific biomart column values to the ones from the original OTP database
                    for index, row in otp_act.iterrows():
                        S1 = otp_targets.loc[otp_targets['ensembl_id']==row['mID']]
                        if len(S1) > 0:
                            otp_act.loc[index,'entrez'] = S1['entrez'].iloc[0]
                            otp_act.loc[index,'gene_type'] = S1['gene_type'].iloc[0]
                            otp_act.loc[index,'hgnc_symbol'] = S1['hgnc_symbol'].iloc[0]
                            otp_act.loc[index,'description'] = S1['description'].iloc[0]
                        else:
                            otp_act.loc[index,'entrez'] = None
                            otp_act.loc[index,'gene_type'] = None
                            otp_act.loc[index,'hgnc_symbol'] = None
                            otp_act.loc[index,'description'] = None
                            Des='Failed to convert the gene ID'
                    otp_act['datasource'] = 'OTP'
                    otp_act['pchembl_value'] = np.nan
                    statement='completed'
                else:
                    statement='Filter reduced interactions to 0'
            else:
                statement='No interaction data'
        else:
            statement = 'No ChEMBL ids found'
            
        return otp_act, statement, otp_raw


    def compounds(self, input_protein: pd.DataFrame):
        """
        Retrieves compounds from OTP database interacting with proteins passed as input.

        Steps
        -----
        - Retrieves gene data from OTP with its API to find compounds interacting with gene \\
        Constraints:
            - None
        - Uses Pubchempy to retrieve the compound info (with additional data from chembl) to return, 
        searching with DrugId or with inchi (obtained from chembl API) if the first search is unsuccessful.
        
        Parameters
        ----------
        input_protein : DataFrame
            Dataframe of input proteins from which interacting compound are found

        Returns
        -------
        DataFrame
            Dataframe of interacting compounds, containing the following values: \\
            inchi, inchikey, isomeric_smiles, iupac_name, datasource (OTP), pchembl_value, notes (nan)
        """
            
        columns = ['inchi','inchikey','isomeric_smiles','iupac_name','datasource','pchembl_value']
        # Create an empty DataFrame with the specified columns
        otp_c1 = pd.DataFrame(columns=columns)
        otp_raw = pd.DataFrame()
        # Drop duplicated and null values of ensembl gene id
        input_protein = input_protein.dropna(subset=['ensembl_gene_id'])
        input_protein = input_protein.drop_duplicates(subset='ensembl_gene_id').reset_index(drop=True)    
        # Check if there are any input proteins remaining  
        if len(input_protein) > 0:
            # Using ensembl gene id to search gene in otp
            input_protein_id = input_protein['ensembl_gene_id'][0]
            
            otp_raw = self.data_manager.retrieve_raw_data('proteins', input_protein_id)

            pc = PubChemServer()
            if len(otp_raw) > 0:    
                # Perform search using drugId and inchis if formers miss
                otp_c1 = self._pubchem_search_chembl(otp_raw, 'drugId', columns, pc)

                if len(otp_c1) > 0:
                    # Store into dataframe add additional data
                    otp_c1['datasource'] = 'OTP'
                    otp_c1['pchembl_value'] = np.nan
                    otp_c1['notes'] = np.nan
                    statement = 'completed'
                else:
                    statement = 'Compounds not found using Pubchem and ChEMBL'
            else:
                statement = 'No interaction data'
        else:
            statement = 'Input protein doesn\'t contain ensembl id'

        return otp_c1, statement, otp_raw
    

    def _retrieve_bib_compounds(self, chembl_ids):
        # Set base URL of GraphQL API endpoint
        base_url = "https://api.platform.opentargets.org/api/v4/graphql"
        # Note: bibliography of OTP only looks for concurrence of IDs in abstracts, it does not look for associations
        otp_bib = pd.DataFrame(columns=['Chembl','PMID','mID','mLabel','mType','strength'])
        bib_list = []

        for ChEMBLID in chembl_ids:          
            # Set variables object of arguments for GraphQL API
            variables = {"ChEMBL ID": ChEMBLID} 
            # Set the Query for the GraphQL API to get the OTP bibliography
            # This inputs the ChEMBLID variable into the query_string at the %s spot
            query_string="""
            query {
                drug(chemblId: "%s"){
                    literatureOcurrences{
                        rows{
                            pmid
                            sentences{
                                matches{
                                    mappedId
                                    matchedLabel
                                    matchedType
                                }
                            }
                        }
                    }
                }
            }
            """ % ChEMBLID 
            
            while True:
                try:
                    # Perform request
                    dat=requests.post(base_url,json={"query": query_string,"variables": variables}) 
                    # print(dat.status_code) #200 code = Operating Normally
                    # Transform API response from JSON to dictionary
                    res=json.loads(dat.text) 
                    break
                except json.JSONDecodeError as e:
                    print(f"Error: {e}. Retrying in 10 seconds...")
                    time.sleep(10)
            
            # Check if at least one interaction has been found
            if res['data']['drug'] is not None:
                    for row in res['data']['drug']['literatureOcurrences']['rows']:
                        for sentence in row['sentences']:
                            for match in sentence['matches']:
                                try:
                                    # Add protein to DataFrame
                                    bib_list.append({
                                        'Chembl': ChEMBLID,
                                        'PMID': row['pmid'],
                                        'mID': match['mappedId'],
                                        'mLabel': match['matchedLabel'],
                                        'mType': match['matchedType'],
                                        'strength': 'weak'
                                    })
                                except:
                                    continue
        # Create DataFrame
        otp_bib = pd.concat([otp_bib, pd.DataFrame(bib_list)])
        otp_bib = otp_bib.drop_duplicates().reset_index(drop=True)

        return otp_bib
    
    def _retrieve_met_compounds(self, chembl_ids):
        # Set base URL of GraphQL API endpoint
        base_url = "https://api.platform.opentargets.org/api/v4/graphql"

        otp_met = pd.DataFrame(columns=['Chembl','PMID','mID','mLabel','mType','strength'])
        met_list = []

        for ChEMBLID in chembl_ids:
            # Set variables object of arguments for GraphQL API
            variables={"ChEMBL ID": ChEMBLID} 
            # Set the Query for the GraphQL API to get the OTP bibliography
            # This inputs the ChEMBLID variable into the query_string at the %s spot
            query_string="""
            query {
                drug(chemblId: "%s"){
                    id
                    mechanismsOfAction{
                        rows{
                            mechanismOfAction
                            targetName
                                targets{
                                    id
                                    approvedSymbol
                                }
                                references {
                                    source
                                    urls
                                }
                        }
                    }
                }
            }
            """ % ChEMBLID 

            while True:
                try:
                    # Perform request
                    dat=requests.post(base_url,json={"query": query_string,"variables": variables}) 
                    #print(dat.status_code) #200 code = Operating Normally
                    # Transform API response from JSON to dictionary
                    res=json.loads(dat.text) 
                    break
                except json.JSONDecodeError as e:
                    print(f"Error: {e}. Retrying in 10 seconds...")
                    time.sleep(10)

            # Check if a matching interaction has been found
            if res['data']['drug'] is not None:
                for mechanism_row in res['data']['drug']['mechanismsOfAction']['rows']:
                    # Check if the target name exists and if there's at least one target
                    if mechanism_row['targetName'] is not None and len(mechanism_row['targets']) > 0:
                        # Add protein to DataFrame
                        try:
                            met_list.append({
                                'Chembl': ChEMBLID,
                                'mLabel': mechanism_row['targetName'],
                                'mID': mechanism_row['targets'][0]['id'],
                                'PMID': mechanism_row['references'][0]['urls'][0],
                                'mType': 'GP',
                                'strength': 'strong'
                            })
                        except:
                            continue
        # Create DataFrame
        otp_met = pd.concat([otp_met, pd.DataFrame(met_list)])
        otp_met = otp_met.drop_duplicates().reset_index(drop=True)

        return otp_met
    
    def _retrieve_proteins(self, input_protein_id):
        # Create post request
        variables={"ensembl_gene_id": input_protein_id}
        base_url="https://api.platform.opentargets.org/api/v4/graphql"
        # Size 10000 is the maximum number of compounds that can be shown. 
        query = """
        query search{
            target(ensemblId:"%s"){
                approvedSymbol
                id
                knownDrugs(size:10000){ 
                    rows{
                        drugId
                        prefName
                        drugType
                    }
                }
            }
        }
        """% input_protein_id
        dat = requests.post(base_url,json={"query": query,"variables": variables})
        # Store data from API
        res = json.loads(dat.text)
        try:
            # Save relevant API data into dataframe
            otp_raw = pd.DataFrame(res['data']['target']['knownDrugs']['rows'])
            # Remove duplicates of drugId
            otp_raw = otp_raw.drop_duplicates(subset='drugId').reset_index(drop=True)
        except TypeError:
            # There was an error with the data types, return empty dataframe
            otp_raw = pd.DataFrame(columns=['inchi','inchikey','isomeric_smiles','iupac_name','datasource','pchembl_value'])

        return otp_raw

            




from abc import ABC
from ..databases import *
from ..sql_connection import connect_to_mysql
import pandas as pd
import numpy as np

class Pipeline(ABC):

    def __init__(self, execution_mode='local', dbs: dict=None, server_info: dict=None) -> None:
        """
        Parameters
        ----------
        execution_mode : {'local', 'online', 'server'}, default 'local'
            The execution modality of the pipeline:
            - local: retrieves most databases info from local csv files, if available
            - server: retrieves most databases info from MySQL server

        dbs : dict
            Dictionary containing all the database stored into DataFrame objects

        server_info : dict
            Dictionary containing the configuration info to connect to the SQL server \\
            Example:
            {
                "host": "127.0.0.1",
                "user": "user",
                "password": "password",
                "database": "cpie",
            }
        """

        self.cnx = None

        if execution_mode not in ['local', 'server']:
            raise ValueError("Invalid execution mode provided. Must be one of the following: \
                             'local', 'server'.")
        
        if execution_mode in ['local']:
            if dbs is None:
                raise ValueError("Databases not specified.")
        
        if execution_mode in ['server']: 
            if server_info is None:
                raise ValueError("Server info not specified.")
            
            dbs = {}
            self.cnx = connect_to_mysql(config=server_info)

            if not self.cnx or not self.cnx.is_connected():
                raise ConnectionError("Couldn't connect to SQL server.")
            
        self.databases = {
            'pc': PubChem(),
            'chembl': ChEMBL(database=dbs.get('chembl', None), connection=self.cnx),
            'bdb': BindingDB(database=dbs.get('bdb', None), connection=self.cnx),
            'stitch': Stitch(database=dbs.get('stitch', None), connection=self.cnx),
            'ctd': CTD(database=dbs.get('ctd', None), connection=self.cnx),
            'dtc': DTC(database=dbs.get('dtc', None), connection=self.cnx),
            'otp': OTP(),
            'dc': DrugCentral(database=dbs.get('dc', None), connection=self.cnx),
            'db': DB(database=dbs.get('db', None), connection=self.cnx)
        }

        self.database_args = {}

        self.sources = ['PubChem', 'ChEMBL', 'BindingDB', 'Stitch', 'CTD', 'DTC', 'OTP', 'DrugCentral', 'DrugBank']


    def _aggregate_pchembl(self, data: pd.DataFrame, index, comp: pd.DataFrame):
        # Calculate the pchembl value for the target, if none, states so
        pchembl = pd.to_numeric(comp['pchembl_value'], errors='coerce').dropna().unique()
        if len(pchembl) > 0:
            data.loc[index,'pchembl_count'] = len(pchembl)
            data.loc[index,'ave_pchembl'] = np.mean(pchembl)
            data.loc[index,'std_pchembl'] = np.std(pchembl)
        else:
            data.loc[index,'pchembl_count'] = 0
            data.loc[index,'ave_pchembl'] = 'Sources do not provide activity data'
            data.loc[index,'std_pchembl'] = 'Sources do not provide activity data'

        return data

    



from .BindingDB import BindingDB
from .ChEMBL import ChEMBL
from .CTD import CTD
from .Database import Database
from .DrugBank import DB
from .DrugCentral import DrugCentral
from .DTC import DTC
from .OTP import OTP
from .PubChem import PubChem
from .Stitch import Stitch

__all__ = ['BindingDB', 'ChEMBL', 'CTD', 'Database', 'DB', 'DrugCentral', 'DTC', 'OTP', 'PubChem', 'Stitch']


import biomart
import pandas as pd
import io
from ..utils.helper import *

class BiomartServer(metaclass=Singleton):

    def __init__(self) -> None:
        self.mirrors = ['useast', 'www', 'asia']
        self._connect_to_biomart()
    
    def _connect_to_biomart(self):
        for mirror in self.mirrors:
            try:
                self.server = biomart.BiomartServer(f"http://{mirror}.ensembl.org/biomart")
                self.dataset = self.server.datasets['hsapiens_gene_ensembl'] # Homo Sapiens
                return 
            except:
                continue
        raise ConnectionError('Couldn\'t connect to biomart')

    def search(self, input_type, input_id, attributes, column_names):
        
        try:
            bm = self.dataset.search({
            'filters': {
                input_type: input_id
            },
            'attributes': attributes
            })

            targets = pd.read_csv(io.StringIO(bm.text), sep='\t', header=None, names=column_names)
        except:
            targets = pd.DataFrame(columns=column_names)

        return targets
    

    def subset_search(self, input_type, input_ids, attributes, column_names):

        targets = pd.DataFrame(columns=column_names)
        if len(input_ids) > 250:
            for start, end in generate_subsets(len(input_ids), 250):
                subset = input_ids[start:end]
                sub_targets = self.search(input_type, subset, attributes, column_names)
                targets = pd.concat([targets, sub_targets])
            
        else:
            sub_targets = self.search(input_type, input_ids, attributes, column_names)
            targets = pd.concat([targets, sub_targets])
        
        targets = targets.reset_index(drop=True)
        return targets



from abc import ABC, abstractmethod
from ..servers.ChEMBLServer import ChEMBLServer as chembl
from ..servers.PubchemServer import PubChemServer
from ..utils.helper import *
import pandas as pd
import time

class Database(ABC):

    @abstractmethod
    def interactions(self, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def compounds(self, *args, **kwargs):
        raise NotImplementedError
    
    def _pubchem_search_cid(self, db_act: pd.DataFrame, columns: list, pc: PubChemServer):
        # Drop null values of CID
        db_act = db_act.dropna(subset=['CID'])
        cids = db_act['CID'].astype(int).unique()
        # Create a list from cids to make batch search using Pubchempy
        cids = [str(x) for x in cids if 'error' not in str(x) and x != 0]
        db_comps = pd.DataFrame(columns=columns)
        selected_columns = pc.get_columns(columns[:-2])
        if len(cids) > 0:
            compounds = pd.DataFrame()
            # Divide the list into batches of 1000 ids
            if len(cids) > 1000:
                for start, end in generate_subsets(len(cids), 1000):
                    subset = cids[start:end]
                    comps = pc.get_compounds(subset, selected_columns, namespace='cid')
                    compounds = pd.concat([compounds, comps])
                    time.sleep(0.5)
            else:
                compounds = pc.get_compounds(cids, selected_columns, namespace='cid')
            db_comps = compounds.rename(columns=pc.properties)

        return db_comps

    def _pubchem_search_chembl(self, db_act: pd.DataFrame, id_name: str, columns: list, pc: PubChemServer):
        pubchem_chembl = []
        existing_id = []
        inchis = []
        chembl_ids = []
        pubchem_inchi = []
        missing_id = []
        db_comps = pd.DataFrame(columns=columns) 
        selected_columns = pc.get_columns(columns[:-2])
        # Perform search of chembl id
        for ids in db_act[id_name].unique():
            try:
                # Retrieve compound using pubchempy
                compound = pc.get_compounds(ids, selected_columns, namespace='name')
                # Check if exactly one compound has been found
                if len(compound) == 1:
                    pubchem_chembl.append(compound)
                    existing_id.append(ids)
                else:
                    missing_id.append(ids)
            except:
                missing_id.append(ids)
            time.sleep(0.5)

        db_chembl = pd.DataFrame()
        # Store Pubchempy output to dataframe
        if len(pubchem_chembl) > 0:
            db_chembl = pd.concat(pubchem_chembl)
            db_chembl = db_chembl.rename(columns=pc.properties)
            db_chembl['id'] = existing_id
            db_comps = pd.concat([db_comps, db_chembl])

        # Utilize chembl molecule API for failed Pubchem ids
        if len(missing_id) > 0:
            chembl.molecule_search(missing_id, inchis, chembl_ids)
        # Search the inchi for the compound that couldn't be searched by chembl id
        if len(inchis) > 0:
            existing_id = []
            for ids in zip(inchis, chembl_ids):
                try:
                    # Use Pubchempy to find compounds based on inchi
                    pubchem_inchi.append(pc.get_compounds(ids[0], selected_columns, namespace='inchi'))
                    existing_id.append(ids[1])
                except:
                    None
                time.sleep(0.5)
        # Check if at least one compound has been found using Pubchempy with inchi
        if len(pubchem_inchi) > 0:
            db_inchi = pd.concat(pubchem_inchi)
            db_inchi = db_inchi.rename(columns=pc.properties)
            db_inchi['id'] = existing_id
            db_comps = pd.concat([db_comps, db_inchi])

        return db_comps





import requests
import time
import pandas as pd
import numpy as np
from ..servers.BiomartServer import BiomartServer
from ..servers.PubchemServer import PubChemServer
from ..data_manager import *
from ..utils.helper import generate_subsets
from .Database import Database

class PubChem(Database):

    def __init__(self, connection=None, database=None):
        if connection is not None or database is not None:
            raise ValueError('No SQL connection or database expected for Pubchem')
        funcs = {
            'comps' : self._retrieve_compounds,
            'targets' : self._retrieve_targets,
            'proteins' : self._retrieve_proteins
        }
        self.data_manager = APIManager(funcs)

    def _filter_database(self, pubchem_raw: pd.DataFrame, identifier: str, pChEMBL_thres: float):
        
        # Drop duplicates, invalid and empty values of identifier and Activity Values
        pubchem_filt = pubchem_raw.dropna(subset=[identifier, 'Activity Value [uM]'])
        pubchem_filt = pubchem_filt[(~pubchem_filt[identifier].eq('')) & (~pubchem_filt['Activity Value [uM]'].eq(''))]
        # Remove Unspecified, Inactive and Inconclusive activities
        pubchem_filt = pubchem_filt[(~pubchem_filt['Activity Outcome'].isin(['Unspecified', 'Inactive', 'Inconclusive'])) &
                                    # Limit measures to Ki, Kd, IC50, EC50, Potency             
                                    (pubchem_filt['Activity Name']!='CD')]
        # Remove duplicates with same identifier and activity value
        pubchem_filt = pubchem_filt.drop_duplicates(subset=[identifier, 'Activity Value [uM]']).copy()
        # Remove non-numeric characters from the Activity Value (things like > and <)
        pubchem_filt['Activity Value [uM]'] = pubchem_filt['Activity Value [uM]'].str.replace(r'[^0-9\.]','',regex=True)
        # Convert strings into numeric values 
        pubchem_filt['Activity Value [uM]'] = pubchem_filt['Activity Value [uM]'].apply(pd.to_numeric,errors='coerce')
        # Generate pChEMBL values 
        pubchem_filt['pchembl_value'] = -np.log10((pubchem_filt['Activity Value [uM]']/1e6)) 
        # Filter interactions by pChEMBL value
        pubchem_filt=pubchem_filt.loc[pubchem_filt['pchembl_value'] > pChEMBL_thres].reset_index(drop=True) 

        return pubchem_filt

    def interactions(self, input_comp: pd.DataFrame, pChEMBL_thres: float=0):
        """
        Retrieves proteins from PubChem API interacting with compound passed as input.

        Steps
        -----
        - Finds information of input compound using PubChem API and filters the resulting DataFrame \\
        Constraints:
            - Only entries with geneids
            - No Unspecified, Inactive or Inconclusive activities
            - Only Ki, Kd, IC50, EC50, Potency measures
        - Computes the pChembl value and identifies target proteins' taxonomy id, searching with gene ID \\
        Constraint:
            - Only Homo Sapiens genes
        - Uses Biomart to obtain proteins info (modified with data from first API search) to return,
        searching with Entrez ID.
        
        Parameters
        ----------
        input_comp : DataFrame
            Dataframe of input compounds from which interacting proteins are found
        pChEMBL_thres : float
            minimum pChEMBL value necessary for interaction to be considered valid

        Returns
        -------
        DataFrame
            Dataframe of interacting proteins, containing the following values:
            - entrez 
            - gene_type 
            - hgnc_symbol 
            - description 
            - datasource (PubChem) 
            - pchembl_value

        String
            A statement string describing the outcome of the database search 
        DataFrame
            Raw Dataframe containing all PubChem info about the input compound
        """
        
        columns = ['entrez','gene_type','hgnc_symbol','description','pchembl_value','datasource']
        # Create an empty DataFrame with the specified columns
        pubchem_act=pd.DataFrame(columns=columns)
        # Create an empty Pubchem activity DataFrame
        pubchem_raw=pd.DataFrame()
        # Drop null values of CID, to make sure the first line of the dataframe has the CID
        input_comp = input_comp.dropna(subset=['cid']).reset_index(drop=True)
        # Check if there are any input compounds remaining  
        if len(input_comp) > 0:
            # Select compound from Pubchem based on CID
            input_comp_id = input_comp['cid'][0]
            
            pubchem_raw = self.data_manager.retrieve_raw_data('comps', input_comp_id)

            if len(pubchem_raw) > 0:
                # Filter Pubchem for high quality interactions and species
                pubchem_act = self._filter_database(pubchem_raw, 'Target GeneID', pChEMBL_thres)
                
                # Find remaining genes
                gene_list = list(pubchem_act['Target GeneID'].unique()) 
            
                gene_tax = self.data_manager.retrieve_raw_data('targets', gene_list)

                # Merge tax_id information into pubchem_act DataFrame
                pubchem_act = pubchem_act.merge(gene_tax, left_on='Target GeneID', right_on='GeneID', how='left')

                # Filter for rows where tax_id is 9606 (human)
                pubchem_act = pubchem_act.loc[pubchem_act['TaxonomyID']==9606].reset_index(drop=True) 
                
                # Check if there is at least one interaction remaining
                if len(pubchem_act) > 0:
                    # Unify gene identifiers by Converting PubChem with biomart
                    ensembl = BiomartServer()
                    # PubChem uses Entrez IDs
                    input_type = 'entrezgene_id' 
                    attributes = ['entrezgene_id', 'gene_biotype', 'hgnc_symbol', 'description']
                    names = ['entrez','gene_type','hgnc_symbol','description']

                    pubchem_targets = pd.DataFrame(columns=names)
                    
                    genelist = list(pubchem_act['Target GeneID'])
                    # list of genes that need conversion
                    input_genes = [int(i) for i in genelist] 

                    pubchem_targets = ensembl.subset_search(input_type, input_genes, attributes, names)

                    # For each compound, assign specific biomart column values to the ones from the original CTD database    
                    for index, row in pubchem_act.iterrows():
                        S1 = pubchem_targets.loc[pubchem_targets['entrez']==int(row['Target GeneID'])]
                        if len(S1) > 0:
                            pubchem_act.loc[index,'entrez'] = S1['entrez'].iloc[0]
                            pubchem_act.loc[index,'gene_type'] = S1['gene_type'].iloc[0]
                            pubchem_act.loc[index,'hgnc_symbol'] = S1['hgnc_symbol'].iloc[0]
                            pubchem_act.loc[index,'description'] = S1['description'].iloc[0]
                        else:
                            pubchem_act.loc[index,'entrez'] = None
                            pubchem_act.loc[index,'gene_type'] = None
                            pubchem_act.loc[index,'hgnc_symbol'] = None
                            pubchem_act.loc[index,'description'] = None
                            Des='Failed to convert the gene ID'
                    pubchem_act['datasource']='PubChem'
                    statement = 'completed'
                else:
                    statement = 'Filter reduced interactions to 0'
            else:
                statement = 'No interaction data'
        else:
            statement = 'Input compound doesn\'t contain CID'

        return pubchem_act, statement, pubchem_raw
    

    def compounds(self, input_protein: pd.DataFrame, pChEMBL_thres: float=0):
        """
        Retrieves compounds from PubChem database interacting with proteins passed as input.

        Steps
        -----
        - Retrieves gene data from PubChem PUG REST API to find compounds interacting with gene: \\
        Constraints:
            - Activity value = Active
        - Uses Pubchempy to obtain compounds info to return, searching with CID and merges information together.
        
        Parameters
        ----------
        input_protein : DataFrame
            Dataframe of input proteins from which interacting compound are found
        pChEMBL_thres : float
            minimum pChEMBL value necessary for interaction to be considered valid

        Returns
        -------
        DataFrame
            Dataframe of interacting compounds, containing the following values:
            - inchi
            - inchikey
            - isomeric smiles
            - iupac name
            - datasource (PubChem)
            - pchembl value (NaN)
            - notes (NaN)
        """

        columns = ['inchi','inchikey','isomeric_smiles','iupac_name','datasource','pchembl_value']
        # Create an empty DataFrame with the specified columns
        pubchem_c1 = pd.DataFrame(columns=columns)
        # Create dataframe from pubchem information
        pubchem_raw = pd.DataFrame()
        # Drop duplicated and null values of entrez
        input_protein = input_protein.dropna(subset=['entrez'])
        input_protein = input_protein.drop_duplicates(subset='entrez').reset_index(drop=True)
        # Check if there are any input proteins remaining         
        if len(input_protein) > 0:
            # Using entrez to search gene in pubchem
            input_protein_id = input_protein['entrez'][0]

            pubchem_raw = self.data_manager.retrieve_raw_data('proteins', input_protein_id)

            if len(pubchem_raw) > 0:
                # Filter Pubchem for high quality interactions and species
                pubchem_act = self._filter_database(pubchem_raw, 'CID', pChEMBL_thres)

                # Check if there are any compounds remaining
                if len(pubchem_act):
                    pc = PubChemServer()

                    # Retrieve compounds using cids
                    compounds = self._pubchem_search_cid(pubchem_act, columns, pc)

                    if len(compounds) > 0:
                        pubchem_act['CID'] = pubchem_act['CID'].astype(int)
                        pubchem_info = pubchem_act[['CID', 'Target Accession', 'pchembl_value']].\
                                        rename(columns={'Target Accession': 'Target Uniprot'}).drop_duplicates()
                        
                        # Add additional values from original request
                        pubchem_c1 = pd.merge(compounds, pubchem_info, on='CID', how='left')
                        pubchem_c1['datasource'] = 'PubChem'
                        pubchem_c1['notes'] = np.nan
                        statement = 'completed'
                    else:
                        statement = 'Compounds not found using PubChem'
                else:
                    statement = 'Filter reduced interactions to 0'
            else:
                statement = 'No interaction data'
        else:
            statement = 'Input protein doesn\'t contain entrez'
            
        return pubchem_c1, statement, pubchem_raw
    
    
    def _retrieve_compounds(self, input_comp_id):

        pubchem_raw=pd.DataFrame()

        if input_comp_id == '':
            Des='No Data'
        else:
            # Use pug rest to get the information from PubChem
            url='https://pubchem.ncbi.nlm.nih.gov/rest/pug/compound/cid/%s/assaysummary/JSON'%(str(input_comp_id))
            response=requests.get(url)
            time.sleep(0.5) 
            data=response.json()
            # Check if a match has been found
            try:
                Des=data['Table']
            except:
                Des='No Data'
        
        if Des!='No Data':
            # Convert API data into a dataframe
            raw_columns = data['Table']['Columns']['Column']
            rows = [row['Cell'] for row in data['Table']['Row']]
        
            pubchem_raw = pd.DataFrame(rows, columns=raw_columns)

        return pubchem_raw
    
    def _retrieve_targets(self, gene_list):
        # Use the PUG REST Pubchem API to assign the tax_id for each gene
        url='https://pubchem.ncbi.nlm.nih.gov/rest/pug/gene/geneid/summary/JSON'
        # Prepare the request payload
        if len(gene_list) > 1000:
            gene_tax = pd.DataFrame(columns=['GeneID', 'TaxonomyID'])
            for start, end in generate_subsets(len(gene_list), 1000):
                subset = gene_list[start:end]        
                payload = {"geneid": ",".join(map(str, subset))}
                try:
                    response=requests.post(url, data=payload)
                    data = response.json()
                    gene_summary_list = data['GeneSummaries']['GeneSummary']
                    genes = pd.DataFrame(gene_summary_list, columns=['GeneID', 'TaxonomyID'])
                    gene_tax = pd.concat([gene_tax, genes])
                except:
                    continue
                time.sleep(0.5)
        else:
            payload = {"geneid": ",".join(map(str, gene_list))}
            try:
                response=requests.post(url, data=payload)
                data = response.json()
                gene_summary_list = data['GeneSummaries']['GeneSummary']
                gene_tax = pd.DataFrame(gene_summary_list, columns=['GeneID', 'TaxonomyID'])
            except:
                gene_tax = pd.DataFrame(columns=['GeneID', 'TaxonomyID'])
        gene_tax['GeneID'] = gene_tax['GeneID'].astype(str)
        return gene_tax
    
    def _retrieve_proteins(self, input_protein_id):

        pubchem_raw = pd.DataFrame()
        # Use PUG REST to get the information of gene activity from PubChem
        url = 'https://pubchem.ncbi.nlm.nih.gov/rest/pug/gene/geneid/%s/concise/JSON' %(str(input_protein_id))
        response=requests.get(url) 
        # Retrieve and filter only relevant data
        data=response.json()
        # Check if a match has been found
        try:
            Des=data['Table']
        except:
            Des='No Data'

        if Des != 'No Data':
            cols=data['Table']['Columns']['Column']
            rows=[]
            for row in data['Table']['Row']:
                rows.append(row['Cell'])
            # Create dataframe from pubchem information
            pubchem_raw = pd.DataFrame(rows, columns=cols)

        return pubchem_raw
            